# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import googleapps


@googleapps.jsonbase
class ParagraphElement(googleapps.JsonBase):
    json_properties = ['page_break', 'column_break', 'footnote_reference', 'horizontal_rule', 'equation', 'inline_object_element', 'auto_text', 'text_run']

    def set_defaults(self):
        self._text_style = None
        self._inline_object = None

    @property
    def element_type(self):
        if self.text_run:
            return('text')

        for field in ('page_break', 'column_break', 'footnote_reference', 'horizontal_rule', 'equation', 'inline_object_element'):
            if getattr(self, field):
                return(field)

        raise Exception("Failed to determine type")

    @property
    def text(self):
        if self.text_run:
            return(self.text_run['content'])

    @property
    def text_style(self):
        if not self._text_style:
            if self.text_run:
                self._text_style = TextStyle(self.text_run['textStyle'], self)
            else:
                self._text_style = TextStyle({}, self)
        return(self._text_style)

    @property
    def inline_object_element_id(self):
        if self.element_type != 'inline_object_element':
            return
        return (self.inline_object_element['inlineObjectId'])

    @property
    def inline_object(self):
        if not self._inline_object:
            oid = self.inline_object_element_id
            if not oid:
                return
            self._inline_object = self.parent.parent.inline_objects[oid]
        return(self._inline_object)




@googleapps.jsonbase
class TextStyle(googleapps.JsonBase):

    json_properties=['bold', 'italic', 'foreground_color', 'font_size', 'weighted_font_family']

    @property
    def link(self):
        if 'link' in self.json:
            return(self.json['link']['url'])

    @property
    def font_family(self):
        if self.weighted_font_family:
            return(self.weighted_font_family['fontFamily'])


    @property
    def font_weight(self):
        if self.weighted_font_family:
            return(self.weighted_font_family['weight'])



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import re
import warnings
from pprint import pprint

import googleapps
#from docx.enum import style

class Border(object):

    border_styles = ('STYLE_UNSPECIFIED','DOTTED','DASHED','SOLID','SOLID_MEDIUM','SOLID_THICK','NONE','DOUBLE')

    def __init__(self, style=None, color=None):
        self.style = style
        self.color = color

    @property
    def style(self): return(self._style)

    @style.setter
    def style(self, style):
        if style and style.upper() not in self.border_styles:
            raise Exception("Unrecognized border style " + style)
        self._style = style

    @property
    def color(self): return(self._color)

    @color.setter
    def color(self, color): self._color = color

    def to_json(self):
        ref = {}
        if self.style:
            ref['style'] = self.style.upper()
        if self.color:
            ref['color'] = Format.color_to_google_color(self.color)
        return ref

class CellBorder(object):

    def __init__(self, style=None, color=None):
        self._top = Border(style, color)
        self._bottom = Border(style, color)
        self._left = Border(style, color)
        self._right = Border(style, color)

    @property
    def top(self):return(self._top)
    @property
    def bottom(self):return(self._bottom)
    @property
    def left(self):return(self._left)
    @property
    def right(self):return(self._right)

    @property
    def sides(self):
        return('top', 'bottom', 'left', 'right')

    def _propogate_to_sides(self, attrib, value):
        for side in self.sides:
            setattr(getattr(self, side), attrib, value)

    @property
    def colors(self):
        return([getattr(getattr(self, side), 'color') for side in self.sides])

    @colors.setter
    def colors(self, value):
        self._propogate_to_sides('color', value)

    @property
    def styles(self):
        return([getattr(getattr(self, side), 'style') for side in self.sides])

    @styles.setter
    def styles(self, value):
        self._propogate_to_sides('style', value)

    def to_json(self):
        ref = {}
        for prop in self.sides:
            v = getattr(self, prop).to_json()
            if v:
                ref[prop] = v
        return(ref)

class Format(object):
    _number_format_type = None
    _number_format_pattern = None
    _background = None
    _foreground = None
    _padding = None
    _alignment = None
    _vertical_alignment = None
    _text_wrap = None
    _text_direction = None
    _text_format = None
    _text_rotation = None
    _hyperlink = None
    _border = None
    _align = None
    _vertical_align = None
    _text_wrap = None
    _rotation = None
    _font_name = None
    _fond_size = None
    _bold = None
    _italic = None
    _strikethrough = None
    _underline = None
    _font_size = None



    number_format_types = ('NUMBER_FORMAT_TYPE_UNSPECIFIED', 'TEXT', 'NUMBER', 'PERCENT', 'CURRENCY', 'DATE', 'TIME', 'DATE_TIME', 'SCIENTIFIC')
    color_re = re.compile(r'#?([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})', re.I)

    def __init__(self, **kwargs):
        for arg in kwargs:
            if not hasattr(self, arg):
                raise Exception(type(self).__module__ + " has no property for " + str(arg))
            setattr(self, arg, kwargs[arg])

    @classmethod
    def color_json(cls, r, g, b, alpha=None):
        return ({ 'red':r, 'green': g, 'blue': b, 'alpha': alpha })

    @classmethod
    def color_8bit_to_google(cls, r, g, b, alpha=None):
        return ({ 'red':r/255, 'green': g/255, 'blue': b/255, 'alpha': alpha })

    @classmethod
    def color_to_google_color(cls, color):
        # { 'red':r, 'green': g, 'blue': b, 'alpha': alpha }
        # [ r, g, b ]
        # 0x02
        # 'red', 'yellow'
        # #ff1144
        # FF33FA
        if color is None:
            return(None)

        # if dict, assume it's already done
        if type(color) is dict:
            return(color)

        # if list, assume a list of numbers. But is it 8 bit RGB, or Google RGB? Find out!
        if type(color) is list:
            if (color[0] > 1 or color[1] > 1 or color[0] > 1) or (color[0] == color[1] == color[2] == 0):
                return(cls.color_8bit_to_google(*color))
            else:
                # google colors! in the range of 0-1
                return(cls.color_json(*color))

        # so, it's a scalar
        if type(color) is int:
                raise Exception("Have not implemented Excel's numerically indexed color format yet")

        # html color or color name?
        m = cls.color_re.match(color)
        if not m:
            # Not a hex value, hopefully it is an actual color. Maybe we should restrict by enumeration?
            return(color)

        return( cls.color_8bit_to_google(int(m.group(1), 16), int(m.group(2), 16), int(m.group(3), 16) ))


    horizontal_alignments = ('CENTER', 'RIGHT', 'LEFT')
    vertical_alignments = ('TOP', 'MIDDLE', 'BOTTOM')

    def fieldmask_fields(self, json=None):
        fields = []
        if not json:
            json = self.cellformat_json()

        for field in json:
            if type(json[field]) is dict:
                subfields = self.fieldmask_fields(json[field])
                fields.extend([ '.'.join([field, sf]) for sf in subfields ])
            else:
                fields.append(field)
        return(fields)

    def cellformat_json(self):
        json = {}

        #--------------------------------------------------------
        # numberFormat
        if self.number_format_type:
            if type(self.number_format_type) is int: # this is a hex value
                raise Exception("Have not implemented Excel's default formats yet")
                #json['numberFormat'] = {'format': self.number_format}
            else:
                json['numberFormat'] = {'type': self.number_format_type}
            if self.number_format_pattern:
                json['numberFormat']['pattern']= self.number_format_pattern

        elif self.number_format_pattern:
                json['numberFormat'] = {'pattern': self.number_format_pattern}

        #--------------------------------------------------------
        # backgroundColor
        if self.background:
            json['backgroundColor'] = self.color_to_google_color(self.background)

        #--------------------------------------------------------
        # border

        if self.border:
            border = self.border.to_json()
            if border:
                json['borders'] = border

        #--------------------------------------------------------
        # padding
        # TODO

       #--------------------------------------------------------
        # Alignment

        if self.align:
            value = self.align.upper()
            if value not in self.horizontal_alignments:
                raise Exception("Unknown alignment format '" + value + "'")
            json["horizontalAlignment"] = value

        if self.vertical_align:
            value = self.vertical_align.upper()
            if value not in self.vertical_alignments:
                raise Exception("Unknown vertical alignment format '" + value + "'")
            json["verticalAlignment"] = value

        if self.text_wrap is not None:
            if type(self.text_wrap) is bool:
                if self.text_wrap:
                    json["wrapStrategy"] = 'WRAP'
                else:
                    json["wrapStrategy"] = 'OVERFLOW_CELL'
            else:
                # allows full support of the google lexicon
                json["wrapStrategy"] = self.text_wrap

        if self.rotation:
            json["textRotation"] = { 'angle': self.rotation }
            # also available is 'vertical': Boolean

        text_format = {}
        if self.foreground is not None:
            text_format['foregroundColor'] = self.color_to_google_color(self.foreground)

        if self.font_name:
            text_format["fontFamily"] = self.font_name

        if self.font_size:
            text_format["fontSize"] = self.font_size

        if self.bold is not None:
            text_format["bold"] = self.bold

        if self.italic is not None:
            text_format["italic"] = self.italic

        if self.strikethrough is not None:
            text_format["strikethrough"] = self.strikethrough

        if self.underline is not None:
            text_format["underline"] = self.underline

        if text_format:
            json["textFormat"] = text_format

        return( json )


        #--------------------------------------------------------

    wrap_strategies = ('OVERFLOW_CELL', 'LEGACY_WRAP', 'WRAP', 'CLIP')

        #--------------------------------------------------------

    @property
    def font_name(self):
        return(self._font_name)

    @font_name.setter
    def font_name(self, value):
        self._font_name = value


    @property
    def font_size(self):
        return(self._font_size)

    @font_size.setter
    def font_size(self, value):
        self._font_size = value


    @property
    def font_color(self):
        return(self._font_color)

    @font_color.setter
    def font_color(self, value):
        self._font_color = value



    @property
    def bold(self):
        return(self._bold)

    @bold.setter
    def bold(self, value):
        self._bold = value


    @property
    def italic(self):
        return(self._italic)

    @italic.setter
    def italic(self, value):
        self._italic = value



    @property
    def underline(self):
        return(self._underline)

    @underline.setter
    def underline(self, value):
        self._underline = value


    @property
    def strikethrough(self):
        return(self._strikethrough)

    @strikethrough.setter
    def strikethrough(self, value):
        self._strikethrough = value


    @property
    def font_script(self):
        return(self._font_script)

    @font_script.setter
    def font_script(self, value):
        self._font_script = value


    @property
    def num_format(self):
        return(self._num_format)

    @num_format.setter
    def num_format(self, value):
        self._num_format = value


    @property
    def locked(self):
        return(self._locked)

    @locked.setter
    def locked(self, value):
        self._locked = value


    @property
    def hidden(self):
        return(self._hidden)

    @hidden.setter
    def hidden(self, value):
        self._hidden = value


    @property
    def align(self):
        return(self._align)

    @align.setter
    def align(self, value):
        self._align = value


    @property
    def vertical_align(self):
        return(self._vertical_align)

    @vertical_align.setter
    def vertical_align(self, value):
        self._vertical_align = value


    @property
    def rotation(self):
        return(self._rotation)

    @rotation.setter
    def rotation(self, value):
        self._rotation = value


    @property
    def text_wrap(self):
        return(self._text_wrap)

    @text_wrap.setter
    def text_wrap(self, value):
        self._text_wrap = value



    @property
    def bg_color(self):
        return(self._bg_color)

    @bg_color.setter
    def bg_color(self, value):
        self._bg_color = value


    @property
    def fg_color(self):
        return(self._fg_color)

    @fg_color.setter
    def fg_color(self, value):
        self._fg_color = value


    def _set_border_attribs(self, border, *args, **kwargs):
        if args:
            if len(args) == 1:
                border.style = args[0]
            else:
                raise Exception("Setting border can only take one arg for the border style")
        for arg, value in kwargs.items():
            setattr(border, arg, value)


    @property
    def bottom_border(self):
        return(self.border.bottom)

    @bottom_border.setter
    def bottom_border(self, *args, **kwargs):
        self._set_border_attribs(self.border.bottom, *args, **kwargs)


    @property
    def top_border(self):
        return(self.border.top)

    @top_border.setter
    def top_border(self,  *args, **kwargs):
        self._set_border_attribs(self.border.top, *args, **kwargs)

    @property
    def right_border(self):
        return(self.border.right)

    @right_border.setter
    def right_border(self,  *args, **kwargs):
        self._set_border_attribs(self.border.right, *args, **kwargs)


    @property
    def left_border(self):
        return(self.border.left)

    @left_border.setter
    def left_border(self,  *args, **kwargs):
        self._set_border_attribs(self.border.left, *args, **kwargs)

    @property
    def border_color(self):
        return(self.border.color)

    @border_color.setter
    def border_color(self, value):
        self.border.color = value

    @property
    def border_style(self):
        return(self.border.styles)

    @border_style.setter
    def border_style(self, value):
        self.border.styles = value

    @property
    def number_format_type(self):
        return(self._number_format_type)

    @number_format_type.setter
    def number_format_type(self, value):

        if value not in self.number_format_types:
            raise Exception("Unknown number format for Google Sheets: " + str(value))
        self._number_format_type = value

    @property
    def number_format_pattern(self):
        return(self._number_format_pattern)

    @number_format_pattern.setter
    def number_format_pattern(self, pattern):
        self._number_format_pattern = pattern

    @property
    def background(self):
        return(self._background)

    @background.setter
    def background(self, color): # color should be a tuple of (r,g,b [a])
        self._background= color

    @property
    def foreground(self):
        return(self._foreground)

    @foreground.setter
    def foreground(self, color): # color should be a tuple of (r,g,b [a])
        self._foreground = color

    @property
    def border(self):
        if not self._border:
            self._border = CellBorder()
        return(self._border)

    @border.setter
    def border(self, cellborder):
        if type(cellborder) is dict:
            self._border = CellBorder(**cellborder)
        else:
            self._border = cellborder

    @property
    def padding(self):
        return(self._padding)

    @padding.setter
    def padding(self, data):
        self._padding = data

    @property
    def alignment(self):
        return(self._alignment)

    @alignment.setter
    def alignment(self, data):
        self._alignment = data

    @property
    def vertical_alignment(self):
        return(self._vertical_alignment)

    @vertical_alignment.setter
    def vertical_alignment(self, data):
        self._vertical_alignment = data

    @property
    def text_wrap(self):
        return(self._text_wrap)

    @text_wrap.setter
    def text_wrap(self, data):
        self._text_wrap = data

    @property
    def text_direction(self):
        return(self._text_direction)

    @text_direction.setter
    def text_direction(self, data):
        self._text_direction = data

    @property
    def text_format(self):
        return(self._text_format)

    @text_format.setter
    def text_format(self, data):
        self._text_format = data

    @property
    def text_rotation(self):
        return(self._text_rotation)

    @text_rotation.setter
    def text_rotation(self, data):
        self._text_rotation = data

    @property
    def hyperlink(self):
        return(self._hyperlink)

    @hyperlink.setter
    def hyperlink(self, data):
        self._hyperlink = data



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import os, os.path
import datetime
import pickle
import re
import shutil
import googleapiclient

import warnings
from pprint import pprint


import googleapps
from .driveitem import DriveItem


class File (DriveItem):

    def get_spreadsheet(self):
        if googleapps.drive.MimeType.google_document_type(mime_type=self.mime_type) == 'spreadsheet':
            return(googleapps.sheets.get(self.id))
        raise Exception("File " + self.name + " (id="+self.id+") is not a Google Apps spreadsheet. Mime Type is " + self.mime_type)
        return

    @property
    def file_type(self):
        return(googleapps.drive.MimeType.file_type(self.mime_type))

    @file_type.setter
    def file_type(self, value):
        self.mime_type = googleapps.drive.MimeType.mime_type(value)

    def is_google_document(self):
        if re.search('google', self.file_type):
            return(True)
        return(False)

    def is_spreadsheet(self):
        return (self.document_type == 'spreadsheet')


    @property
    def document_type(self):
        return(googleapps.drive.MimeType.document_type(mime_type=self.mime_type))

    @property
    def document_mime_type(self):
        return(googleapps.drive.MimeType.convert_document_mime_type(mime_type=self.document_type))

    @property
    def mime_type_extension(self):
        return(googleapps.drive.MimeType.mime_type_extension(self.mime_type))

    def _prepare_upload(self, filepath, mime_type=None, convert=None):

        if not os.path.exists(filepath):
            raise FileNotFoundError("Source file to upload does not exist at path " + str(filepath))

        result = {}

        #now get/load the source file
        if mime_type:
            result['mime_type'] = mime_type
        else:
            match = re.search(r'\.([^\.]+)$', filepath)
            result['mime_type'] = googleapps.drive.MimeType.extension_mime_type(match.group(1))

        # get file times
        stat = os.stat(filepath)
        result['modified_time'] = datetime.datetime.fromtimestamp(stat.st_mtime)
        result['created_time'] = datetime.datetime.fromtimestamp(stat.st_ctime)

        # and the media
        result['media'] = googleapiclient.http.MediaFileUpload(filepath, mimetype=result['mime_type'], resumable=True)

        # Find out if it is eligible to convert to google doc
        if result['mime_type'] and (convert or convert is None):
            doctype = googleapps.drive.MimeType.convert_document_mime_type(mime_type=result['mime_type'])
            if doctype:
                # BY DEFUALT, (convert is None) we convert.
                # We need to do it in this order because we want the orgiinal mime_type set for the
                # MediaFileUpload, but the converted mime_type on the file object
                result['mime_type'] = doctype

        return(result)


    def replace(self, filepath, mime_type=None, convert=None):
        # this is a bit of shorthand as it calls update() at the end to enact the replacement
        # the programmer could do this on their own terms merely by setting media_body as a change variable

        if not self.last_loaded and not self.id:
            raise Exception("replace() called on object, but this is not existing in the Drive yet")

        media_data = self._prepare_upload(filepath, mime_type, convert)

        # error checking if we have available information
        # It's possible we have not actually loaded the file information, in which case
        # we cannot do this error checking
        if self.last_loaded and media_data['mime_type'] != self.mime_type:
            # is this right? What about for convent/unconvert?
            raise Exception("Mime Type " + mime_type + " is derived by the file replace() call, but differs than "+self.mime_type+", which is manually set on the object ")

        # we also update the time variables based on the media data
        for file_time_field in 'modified_time', 'created_time':
            setattr(self, file_time_field, media_data[file_time_field])

        self.changes['media_body'] = media_data['media']

        return(self.update())


    def upload(self, filepath, mime_type=None, convert=None):
        if self.last_loaded:
            raise Exception("upload() is designed for an object that is not in the google drive yet")

        # automatically set the name from the filepath if it doesn't already exist
        if not 'name' in self.changes:
            #TODO: test this
            match = re.search(r'([^\/\\]+)$', filepath)
            if not match or not match.group(1):
                raise Exception("Cannot determine filename from " + filepath)
            self.name = match.group(1)

        # get metadata parameters
        media_data = self._prepare_upload(filepath, mime_type, convert)
        if 'mime_type' not in self.changes:
            self.mime_type = media_data['mime_type']
        elif media_data['mime_type'] != self.mime_type:
            # Is this necessary?
            raise Exception("Mime Type " + mime_type + " is derived by the file upload() call, but differs from "+self.mime_type+", which is manually set on the object ")

        # to determine the correct parameters, we follow this protocol:
        # 1. If the properties were manually set on the object by the programmer prior to the upload() call, that takes precedence over the file properties
        # 2. By default, the file time properties (create, access, modified) are set for the upload

        # this diverges from the replace() method because when upload() is called, it is assumed not to be in the drive yet, and so if the programmer
        # set the times prior to the call to upload(), they should want those to be override times.

        for file_time_field in 'modified_time', 'created_time':
            if file_time_field not in self.changes:
                setattr(self, file_time_field, media_data[file_time_field])

        self.changes['media_body'] = media_data['media']
        return(self.save())



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import re
import warnings
import datetime
import json

from pprint import pprint
import googleapps
import googleapiclient
import time

class DefaultProperty:
    """ An override to the @property decorator that simplifies the property process and allows for a readonly definition.

    Args:
        propertyname (str): The name of the property. This will be set on all objects. It will also create a private instance variable preceded by an underscore where the value is actually stored.
        readonly (bool): Whether this property is readonly. Default is false.
    """

    def __init__(self, propertyname, readonly=False):

        self.property = propertyname
        self.readonly = readonly
        self.attr = '_' + propertyname

    def __get__(self, obj, cls=None):
        """ When accessed, return the current value of the attributed. When called in class contexts, return the property object. """
        if instance is None:
            return(self)
        return( getattr(obj, self.attr ))

    def __set__(self, obj, value):
        """ The function for setting the property, which sets the attribute .

            Raises AttributeError if this is a readonly property.
        """
        if self.readonly:
            raise AttributeError("Attempted to set '" + self.property + "' on "+type(obj).__name__+", which is a read-only attribute.")
        return( setattr(obj, self.attr, value))

    def __delete__(self, obj):
        """ The function for deleting the object.

        Raises AttributeError if this is a readonly property.
        """
        if self.readonly:
            raise AttributeError("Attempted to delete'" + self.property + "' on "+type(obj).__name__+", which is a read-only attribute.")
        return( delattr(obj, self.attr))

class RemoteProperty(DefaultProperty):
    ''' A property class for Python classes that represent remote objects. This is an extension of the property interface with the following:

    - the setter class record_value_change(property, value), which will allow the class to make/queue any request that this change requires.

    - TODO: still trying to figure out what the appropriate functionality for del is
    '''

    def __get__(self, obj, cls=None):
        """ Gets the value. In cases when the value has not been defined, it attempts to commit the object (if permitted) and checks again - ie. perhaps this is a new object and some properties are defined when the object is first saved remotely (like an autogenerated ID or a date created), so the commit is attempted.

        Raises AttributeError if the attribute hasn't previously been defined (or still is not definied if autocommit is permitted and successful), because in these cases it cannot be known what the value actually is.
        """

        if not hasattr(obj, self.attr):
            if obj.commit_policy >= CommitType.defer_commit:
                if obj.debug:
                    print("-- Automatic commit() triggered by call on "+type(obj).__name__ +" object to property " + self.attr + " before it has been fetched")
                obj.commit()
                if not hasattr(obj, self.attr):
                    raise AttributeError(type(obj).__name__ + " property '"+str(self.property)+"' is not available for this object, even after attempting to commit the current state")
            else:
                raise AttributeError(type(obj).__name__ + " property '"+str(self.property)+"' is not available. Have you committed the necessary service requests to get this information from the remote source?")
        return(getattr(obj, self.attr))

    def __set__(self, obj, value):
        """ After setting, calls record_value_change on the object to register the change. """
        result = super().__set__(obj, value)
        obj.record_value_change(self.property)
        return(result)

    def __delete__(self, obj):
        """ deletes the value.
             - if a current value has been set/recorded, that value is cleared. Fully cleared from the queue. It is NOT replaced by "None"
            - we should not communicate this property value to the server on any future commit()
            - we should automatically set the local value on load()
            - hasattr() will return false, and so if the property is requested before the next load() or commit(), the AttributeError will be thrown

        Notes:
            TODO: verify this is correct and desired.

            Delete could work in several different ways.... Which is the most correct functionality for delete?
                a. Delete is remotely-attached: By deleting the property, the call is intended to clear the value on the Remote source.
                b. (current implementation) Delete is locally-attached: The delete should clear the local value only, indicating that load() will overwrite it locally and commit() will *not* send anything about this property to the Remote source
        """
        result = super().__delete__(obj)
        obj.record_value_change(self.property, cleared=True)
        return( result )


def classproperties(class_obj, *args, **kwargs):
    """ This is a decorator for a class that definies a series of properties """

    def set_property(cls, property, definition):
        if hasattr(cls, property):
            initial_value = getattr(cls, property)

        setattr(cls, property, definition)

        if initial_value is not None:
            # TODO: if the class has a value set, should it do the full registration (which might be important for a complicated setter with side effects)
            # or not?  Currently, not. Initial/default values do not trigger any side effects.
            setattr(class_obj, definition.attr, initial_value)

    # by default, a list of properties become DefaultProperty
    if hasattr(class_obj, 'properties') and 'properties' not in kwargs:
        for property in class_obj.properties:
            set_property(class_obj, property, DefaultProperty(property))

    # any other args parameter is a custom-named list of properties that are not "custom properties", so they also become DefaultProperty
    for property_list in args:
        properties = getattr(class_obj, property_list)
        for property in properties:
            set_property(class_obj, property, DefaultProperty(property))

    # any kwargs is a list of "custom properties" and set the list name to the property class they should be defined to.
    # Note that setting the default 'properties' list will be handled here because of the 'not in kwargs' in the default
    # any propety list with a standalone "ro" will be set to readonly, so 'ro_properties', 'props_ro', 'auto_ro_props' all qualify
    for property_list, propdefinition in kwargs.items():
        readonly = re.search(r'(^|\_)ro(\_|$)', property_list)
        for property in getattr(class_obj, property_list):
            setattr(class_obj, property, propdefinition(property, readonly=readonly))

def googleappclient(cls):
    # This is just shorthand for a specific interface for the classproperties
    classproperties(cls, api_properties=RemoteProperty, api_ro_properties=RemoteProperty)
    return(cls)

class ServiceRequest():
    def __init__(self, api, api_version, resource, function, callback, parameters):
        self.api=api
        self.api_version = api_version
        self.resource = resource
        self.function = function
        self.callback = callback
        self.parameters = parameters

    def append(self, *param_path):
        json = param_path[-1]
        ref = self.parameters
        for var in param_path[:-1]:
            ref = ref[var]
        ref.append(json)


import collections
import enum
class CommitType(enum.IntEnum):
    no_commit = 0
    defer_commit = 1 # don't commit without explicit user statement.  load() and save() are exceptions
    entity_commit = 2 # commit all major operations, such as creating or deleting a new entity. Do not update minor property changes
    auto_commit = 3  # commit everything.

class ServiceClient():
    services = googleapps.ServiceManager()
    default_api = None
    default_api_version = None
    default_scope = None
    debug=False
    debug_commit=False

    api_properties = ()
    api_properties = ()
    api_translation = ()

    commit_policy=CommitType.auto_commit
    attempt_recover=True
    retry_count=5
    quotas = ({'timespan':'min', 'count':100})

    def __init__(self, **kwargs):

        self._commit_policy = self.__class__.commit_policy
        for arg in kwargs:
            if arg not in self.api_properties and arg not in self.api_ro_properties:
                raise TypeError("__init__() got unexpected keyword argument '"+arg+"'")
            setattr(self, '_' + arg, kwargs[arg])

        self.clear_queue()

    @classmethod
    def _generate_api_properties(cls):
        for property in cls.api_properties:
                setattr(cls, property, autoproperty(property))

    def _initialize_from_template(self):
        for property in self.new_property_template:
            attr = '_' + property
            if not hasattr(self, attr):
                setattr(self, property, self.new_property_template[property])


    def commit(self):
        # now do everything
        for request in self.request_queue:
            # now parse the parameters
            response = self._send_request(request)

        self.clear_queue()

    def make_request(self, resource, function, callback=None, api=None, api_version=None, **parameters):
        if not api:
            api = self.default_api
        if not api_version:
            api_version = self.default_api_version
        return(self._send_request(ServiceRequest(api, api_version, resource, function, callback, parameters)))

    def add_request(self, resource, function, callback=None, api=None, api_version=None, commit_on=CommitType.entity_commit, **parameters):
        if not api:
            api = self.default_api
        if not api_version:
            api_version = self.default_api_version
        request = ServiceRequest(api, api_version, resource, function, callback, parameters)
        if self.commit_policy >= commit_on:
            self._send_request(request)
        else:
            self.request_queue.append(request)

    def add_update_request(self, *args, commit_on=CommitType.auto_commit, **parameters):
        # same as add_request, except the commit policy is more stringent
        self.add_request(*args, commit_on=commit_on, **parameters)


    @property
    def last_request(self):
        if len(self.request_queue):
            return(self.request_queue[-1])


    def clear_queue(self):
        self.request_queue = []


    @classmethod
    def _prepare_json(cls, parameters, object=None):
        result = {}
        if type(parameters) is not dict:
            return(parameters)
        for key, value in parameters.items():
            if callable(value):
                if object:
                    result[key] = value(object)
                else:
                    raise Exception("JSON parameter " + str(key) + " is set as a deferred calculation, but the prepare_json was called in class context without an object on which to call the function")
            elif type(value) is dict:
                result[key] = cls._prepare_json(value, object=object)
            elif type(value) in (list, tuple):
                result[key] = []
                for v in value:
                    # this is naive. We don't really think v is always going to be json, do we?
                    result[key].append(cls._prepare_json(v, object=object))
            elif type(value) is datetime.datetime:
                result[key] = googleapps.to_rfctime(value)
            else:
                result[key] = value
        return(result)

    def prepare_json(self, parameters):
        return(self.__class__._prepare_json(parameters, object=self))

    @property
    def default_service(self):
        #TODO: we either needto change the class variables to 'api' and 'api_version' or handle instance variables here.
        return(self.get_service(self.default_api, self.default_api_version))

    @classmethod
    def get_service(cls, api=None, api_version=None):
        if api:
            if not api_version and api == cls.api:
                return(cls.services.load_service(cls.default_api, cls.default_api_version, cls.default_scope))

            return(cls.services.load_service(api, api_version))
        else:
            return(cls.services.load_service(cls.default_api, cls.default_api_version, cls.default_scope))


    def _send_request(self, request):
        return(self.__class__.make_class_request(request, object=self, attempt_recover=self.attempt_recover, retry_count=self.retry_count, debug=self.debug))

    @classmethod
    def make_class_request(cls, request, object=None, debug=None, attempt_recover=True, retry_count=4):

        service = cls.get_service(request.api, request.api_version)

        if type(request.resource) is list:
            resource = service
            for res in request.resource:
                resource = getattr(resource, res)()
        else:
            resource = getattr(service, request.resource)()
        function = getattr(resource, request.function)
        # remember we can defer by providing closues.  It's not timet o execute them

        request_json = cls._prepare_json(request.parameters, object=object)
        if debug:
            print("---------------\n " + str(request.resource) + " is about to commit a '"+str(request.function)+"' request:")
            pprint(request_json)
        try:

            response = function(**request_json).execute()
        except googleapiclient.errors.HttpError as he:
            if not attempt_recover:
                raise(he)

            hejson = json.loads(he.content)

            # Other errors that might be added as specific handling
            # "The caller does not have permission"
            #
            #-----
            if he.resp.status in (400,403) or re.search('Invalid (request|JSON|value)', hejson['error']['message']):
                # just propogate the error
                print("Error in service request. Parameters sent in request: ", request.parameters)
                raise
            elif re.search(r'Insufficient Permission: Request had insufficient authentication scopes.', hejson['error']['message']):
                print("The scopes you have authorized are not sufficient enough for the function called. You will need to delete the cached authentication file and reauthorize with appropriate scopes.")
                raise
            elif re.search("Insufficient tokens for quota", hejson['error']['message']):
                print("Quota exceeded. Attempting to sleep for afew minutes and trying again")
                wait_time = 60
            else:
                print("HttpError caught for " + str(request.resource) + " execution of " + str(request.function) + ": " + str(he))
                print("-- Attempting wait and retry -- ")
                wait_time = 1

            counter = retry_count
            while counter:
                time.sleep(wait_time)
                print('..... retrying request (retry='+str(retry_count - counter + 1)+') .....')
                try:
                    response = function(**request_json).execute()
                    #success! Exit loop!
                    # note counter is still > 0
                    break
                except googleapiclient.errors.HttpError as he:
                    if cls.debug:
                        import pdb
                        pdb.set_trace()
                    # decrease the counter, double the wait time
                    counter -= 1
                    wait_time *= 2

            if not counter:
                print("Failed to recover. Exiting.")
                raise

        if debug:
            print("\n***** Response Received: *****")
            pprint(response)
            print("---------------\n")
        if request.callback:
            callback = request.callback
            if debug:
                print("----> Callback provided: " + str(callback))
            callback(response)

        return(response)


    #
    #
    # I think this is deprecated
    #
    #

    def process_api_definition(self, request_function, values, parameters, *nested_parameters):
        request = {}
        keys = list(values.keys())
        for arg in keys:
            if arg not in parameters:
                raise Exception("Unknown parameter '" + arg + "' passed to " + request_function )

            if values[arg] is None:
                warnings.warn("In "+request_function+", parameter '" + arg + "' passed as None, which will be ignored" )
            else:
                request[googleapps.toCamel(arg)] = values[arg]
            del values[arg]


        for nested in nested_parameters:
            for nest, sub_parameters in nested.items():
                subrequest = self.process_api_definition(
                    request_function + ', subfunction' + nest,
                    values,
                    sub_parameters)
                if subrequest:
                    request[googleapps.toCamel(nest)] = subrequest

        if values:
            raise Exception("Unknown parameter-value ( "+str(values)+" ) provided to " + request_function )

        if request:
            return(request)



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import os, os.path
import datetime
import dateutil.parser
import pickle
import re
import shutil
import io
import time
import warnings
from pprint import pprint

import googleapps
from googleapps.serviceclient import CommitType
import googleapiclient

def mime_type_class(mime_type):
    default = googleapps.drive.File

    mime_type_classes = {
        # expand this as more classes are defined for the Drive Item Factory
        'application/vnd.google-apps.file': googleapps.drive.File,
        'application/vnd.google-apps.folder': googleapps.drive.Folder,
        #'application/vnd.google-apps.spreadsheet': googleapps.sheets.Spreadsheet,
    }

    if mime_type in mime_type_classes:
        return(mime_type_classes[mime_type])
    return(default)

@googleapps.serviceclient.googleappclient
class DriveItem(googleapps.ServiceClient):
    default_api = 'drive'
    default_api_version = 'v3'
    cache_directory = 'cache'
    services = googleapps.ServiceManager()

    time_fields = ('trashed_time', 'viewed_by_me_time', 'created_time', 'modified_time', 'modified_by_me_time', 'shared_with_me_time')

    # these properties are automatically generated into properties the googleappclient decorators
    api_ro_properties = ('id','md5_checksum', 'size',' quota_bytes_used', 'head_revision_id',)
    api_properties = ('name', 'mime_type', 'description', 'starred', 'trashed', 'explicitly_trashed', 'trashing_user', 'trashed_time',
                      'version', 'web_content_link', 'web_view_link', 'icon_link', 'has_thumbnail', 'thumbnail_link', 'thumbnail_version',
                      'modified_time', 'modified_by_me', 'modified_by_me_time',
                      'owned_by_me', 'shared', 'shared_with_me_time',
                      'viewed_by_me_time', 'viewers_can_copy_content', 'writers_can_share',
                      'original_filename', 'fulll_file_extension', 'file_extension',
                      # TODO: these should be managed more individaully
                      'parents', 'spaces', 'permissions', 'permission_ids',
                      # owners, permissions,
                      'drive_id'
                      )
    absent_google_fields = ('md5_checksum', 'drive_id')

    def __init__(self, *args, **kwargs):
        # can create a drive item as :
        # DriveItem(id)

        # DriveItem(name=TITLE) - create new file

        # TODO: Verify default behavior of DriveItem(single_param) is that single_param is the name. Not the ID.
        if len(args) == 1:
            kwargs['id'] = args[0]
            args = args[1:]
        elif len(args) > 1:
            raise Exception("googleapps.drive does not provide support for more than one ordered argument")

        super().__init__(**kwargs)
        self.last_loaded = None
        self.changes = {}

    def is_folder(self):
        return(self.__class__.__name__ == 'Folder')

    def is_file(self):
        return(self.__class__.__name__ == 'File')

    @property
    def in_shared_drive(self):
        return(self.drive_id is not None)

    def folders(self):
        return([googleapps.drive.Folder(id) for id in self.parents])

    @property
    def folder(self):
        folders = self.folders()
        if len(folders) > 1:
            self._die('folder()', "but it has multiple parents: '" + "','".join(self.parents) + "'")
        return(folders[0])

    def load(self):
        self.make_request('files', 'get', fileId=self.id, supportsAllDrives=True, fields='*', callback=self.process_request_json)

    def refresh(self):
        # at present this is no different than load
        # but we keep it separate in case we want to abstract it further sometime in the future
        self.load()

    def save(self):
        if self.last_loaded:
            # just update, because it already exists remotely
            self.update()
        else:
            # new entity, so gather all the properties and make the create() request
            parameters = {'body': self._create_json_parameters()}
            if 'media_body' in self.changes:
                parameters['media_body'] = self.changes['media_body']
            # any call to create should update the rest of the properties with the result
            if self.debug:
                print("---------------- SAVING NEW OBJECT USING files().create(): -------------")
                pprint(parameters)
                print("------------------------------------------------------------------------")
            result = self.make_request('files', 'create', fields='*', callback=self.process_request_json, **parameters)

        # okay, regardless of the path above, process the result
        self.changes = {}

    def _create_json_parameters(self):
        parameters = {}
        for property in self.api_properties:
            if hasattr(self, property):
                if property == 'parents':
                    # special processing - get the parent ids
                    # we defer fetching IDs until here because it is possible in non-autocommit situations
                    # that the parent may not have been saved() yet
                    parameters['parents']= [parent if type(parent) is str else parent.id for parent in getattr(self, property)]
                else:
                    parameters[googleapps.toCamel(property)] = getattr(self, property)
        return(parameters)


    #-----------------------------
    #
    # Top-level requests
    #
    #-----------------------------

    def delete(self):
        self.add_request('files', 'delete', fileId=self.id )
        # clear all changes
        self.changes = {}
        self.deleted = True
        return(True)

    def copy(self, name=None, to_folder=None):
        body = {}

        if name:
            body['name'] = name

        if to_folder:
            if type(to_folder) is str:
                body['parents'] = [ to_folder ]
            else:
                body['parents'] = [ to_folder.id ]

        new_obj = self.__class__()
        self.add_request('files', 'copy',
            fileId=self.id,
            body=body,
            fields='*',
            callback=new_obj.process_request_json
        )
        return(new_obj)

    def download(self, filepath=None, mime_type=None, convert_document=True, overwrite=False):
        # by default, export to an io.BytesIO handle object
        # filepath can be a directory or a full path
        # if the directory exists, it will save it as a file with self.name
        # if no directory, it will save as the file
        # If it is a document convert, it will append an extension if it is not part of the filename


        # determine the filename (or not) to write data to
        fullpath = None
        if filepath:
            if os.path.exists(filepath) and os.path.isdir(filepath):
                fullpath = os.path.join(directory, self.name)
            else:
                fullpath = filepath

        # determine the request, which can be a straight download or an export, especially an export of a Google Document
        request = None
        googledoc_convert = None
        if not mime_type and fullpath and self.is_google_document() and convert_document:
            fextmatch = re.search('\.(\w+)$', fullpath)
            if not fextmatch:
                raise Exception("Cannot get extension for filename " + fullpath)
            document_mime_type = googleapps.drive.MimeType.extension_mime_type(fextmatch.group(1))
            request = self.default_service.files().export_media(fileId=self.id, mimeType=document_mime_type)
        elif mime_type:
            request = self.default_service.files().export_media(fileId=self.id, mimeType=mime_type)
        else:
            request = self.default_service.files().get_media(fileId=self.id)


        # open the filehandle, possibly adding the extension if necessary
        if fullpath:
            if googledoc_convert:
                doc_extension = googleapps.drive.MimeType.mime_type_extension(googledoc_convert)
                if not re.search(r"\\." + doc_extension + "$", fullpath, re.I):
                    fullpath += '.' + doc_extension

            if os.path.exists(fullpath) and not overwrite:
                raise Exception("Attempting to overwrite " + fullpath + ", but the overwrite parameter in the download() call is set to False.")
            fh = open(fullpath, 'wb')
        else:
            fh = io.BytesIO()

        # process the download
        downloader= googleapiclient.http.MediaIoBaseDownload(fh, request)
        done = False
        iterations = 0
        at_99 = 0
        while done is False:
            #import pdb
            #pdb.set_trace()
            try:
                status, done = downloader.next_chunk()
            except googleapiclient.errors.HttpError as he:
                print("---- caught exception in chunk download.  waiting for a couple seconds -----")
                print(str(he))
                time.sleep(3)
                status, done = downloader.next_chunk()

            iterations += 1
            if self.debug:
                if (iterations % 20) == 0:
                    print('.')
                else:
                    print('.', end="")
            # we ran into a bug in which the download would never end, so we're going to try to investigate more
            if status.progress() > .99:
                at_99 += 1
                if at_99 > 10:
                    raise Exception('Likely download error. Download at 99% for over 10 iterations.  Investigate?')

        if fullpath:
            fh.close()
            os.utime(fullpath, times=(datetime.datetime.now().timestamp(), self.modified_time.timestamp()))
            return(fullpath)
        else:
            return(fh)

    #-----------------------
    #
    # Property Updates
    #
    #-----------------------

    def rename(self, new_name):
        self.name = new_name

    @classmethod
    def _resolve_ids(cls, objects_or_ids):
        return([id if type(id) is str else id.id for id in objects_or_ids])

    def move(self, to_folder):
        # best to copy the remove_from list so that we avoid conflicts of the array passed by reference
        # TODO: have the 'parents' property return a copy to beging with
        self.add_folder_change(add_to=[to_folder], remove_from=self.parents.copy())

    def remove_from_folder(self, *folders):
        self.add_folder_change(remove_from=folders)

    def add_to_folder(self, *folders):
        self.add_folder_change(add_to=folders)

    #-----------------------
    #
    # Internal Functions
    #
    #-----------------------

    def add_folder_change(self, add_to=None, remove_from=None, allow_commit=True):
        # allow_commit is whether or not to consider this at the top level - if a number of actions will be occurring,
        # passing in allow_commit=False can make it seem like an atomic call
        self.changes.setdefault('folder_changes', [])

        if remove_from:
            folder_ids = self._resolve_ids(remove_from)
            for folder_id in folder_ids:
                try:
                    index = self.parents.index(folder_id)
                    del self.parents[index]
                    self.changes['folder_changes'].append(['remove', folder_id])
                except ValueError as e:
                    raise ValueError("Folder '" + index + "' is not a parent of '" + self.id + "', and so it cannot be removed")

        if add_to:
            folder_ids = self._resolve_ids(add_to)
            for folder_id in folder_ids:
                if folder_id not in self.parents:
                    self.changes['folder_changes'].append(['add', folder_id])
                    self.parents.append(folder_id)


        if allow_commit and self.commit_policy >= CommitType.auto_commit:
            self.update()


    def update(self):

        root_properties = ('keep_revision_forever', 'ocr_language', 'supports_all_drives', 'use_content_as_indexable_text', 'media_body')

        root_parameters = {}
        body_parameters = {}

        for property in self.changes:
            if property == 'folder_changes':
                continue
            elif property in root_properties:
                root_parameters[googleapps.toCamel(property)] = getattr(self, property)
            else:
                body_parameters[googleapps.toCamel(property)] = getattr(self, property)

        # allow for the possibility of both adding and removing parents before committing updates
        if 'folder_changes' in self.changes:
            to_add = {}
            to_remove = {}
            for change in self.changes['folder_changes']:
                if change[0] == 'add':
                    if change[1] in to_remove:
                        del to_remove[change[1]]
                    to_add[change[1]]= True
                else:
                    if change[1] in to_add:
                        del to_add[change[1]]
                    to_remove[change[1]]= True
            if to_add:
                root_parameters['addParents'] = ','.join(to_add)
            if to_remove:
                root_parameters['removeParents'] = ','.join(to_remove)

        # all set, make the request
        self.make_request('files', 'update',
                          fileId=self.id,
                          callback=self.process_request_json,
                          fields='*',
                          body=body_parameters,
                          **root_parameters,
                          )


    def record_value_change(self, property, cleared=False):
        if cleared:
            del self.changes[property]
        else:
            # if autocommit *AND* This is already an existing element, make the change
            # if it does not exist on Drive yet, we don't make the change because
            # it may be happening during an atomic transaction (upload, create, ....)
            # We may want to move toward a more atomic-transaction system at some point
            if self.commit_policy == CommitType.auto_commit and self.last_loaded:
                result = self.make_request('files', 'update',
                                  fileId=self.id,
                                  body={googleapps.toCamel(property): getattr(self, property)},
                                  fields='*',
                                  callback = self.process_request_json
                                  )
            else:
                self.changes[property] = True

    def process_request_json(self, raw_json):
        json = googleapps.unCamel(raw_json)
        if self.debug:
            print("--- Received JSON to Process ----")
            pprint(json)
            print("---------------------------------")

        # first, if this is the fist actuall remote call we've made, set the ID
        if not hasattr(self, 'id'):
            self._id = json['id']

        for key in self.api_properties + self.api_ro_properties:
            if key in json:
                if self.debug:
                    print("\tSet " + key)
                    if hasattr(self, key):
                        current = getattr(self, key)
                        if json[key] != current:
                            print("\t\t**** This value ("+str(json[key])+") is different than the current existing value of : " + str(current))
                # special processing, convert for time fields
                if key in self.time_fields:
                    setattr(self, '_' + key, googleapps.from_rfctime(json[key]))
                else:
                    setattr(self, '_' + key, json[key])
            elif key in self.absent_google_fields:
                if self.debug:
                    print("Setting absent field " + key + " to none")
                # Because these properties will die with a Not Loaded error if they are called,
                # and this would signify they were loaded,
                # we set them as none if they don't exist
                setattr(self, '_' + key, None)
            else:
                if self.debug:
                    print("NOT SETTING absent field " + key)


        #TODO: Verify this is an appropriate decision
        #  BEST AND NEED TO DO: - the timestamp right before the request was made, which is probably safest
        # could be:
        #  same as the later of modified/view time, reported by Google
        #  currently, which is the timestamp this function began running, which is just the worst option but easy for now
        self.last_loaded=datetime.datetime.now()




    def _die(self, function, msg):
        identifying_string = str(function) + " called for Google Drive item "
        if hasattr(self, 'id'):
            if hasattr(self, 'name'):
                identifying_string += self.name + " (id='" +str(self.id)+ "')"
            else:
                identifying_string += "'" +str(self.id)+ "'"
        elif hasattr(self, 'name'):
            identifying_string += "'" +str(self.name)+ "'"
        else:
            pass
        raise Exception(identifying_string + msg)


    @classmethod
    def process_queryvalue(cls, value):
        if value is None:
            return
        if type(value) is str:
            value = re.sub("'", "\\'", value)
            return("'" + value + "'")
        elif type(value) in (datetime.datetime, datetime.date, datetime.time):
            return googleapps.to_rfctime(value)
        elif type(value) is bool:
            if value:
                return('true')
            else:
                return('false')
        return(str(value))


    @classmethod
    def generate_query_string(cls, conjunction='and', **fields):

        q = []
        for field,value in fields.items():
            camelfield = googleapps.toCamel(field)
            if field in ['or', 'and']:
                # disjunction instead of conjunction
                q.append('(' + cls.generate_query_string(conjunction=field, **value) + ')')
            elif type(value) is dict:
                if value['op'] in ('in', 'not in'):
                    # 'in' is reversed
                    q.append(cls.process_queryvalue(value['value']) + ' '+value['op']+' ' + camelfield)
                else:
                    q.append(camelfield + ' ' + value['op'] + ' ' + cls.process_queryvalue(value['value']))
            elif type(value) is list:
                value_strings = []
                for val in value:
                    value_strings.append(camelfield + ' = ' + cls.process_queryvalue(val))
                q.append('(' + ' OR '.join(value_strings) + ')')

            elif value is not None:
                # in some cases (like with the trashed parameter, the user can explicitly be
                # set to None to mean that the default should not be passed in - and in fact
                # nothing should be provided, so we leave the elif to only add a parameter
                # if it is not none)
                q.append(camelfield + ' = ' + cls.process_queryvalue(value))

        if cls.debug:
            print("---- Query String Generated -----")
            print(q)
        if len(q) == 1:
            return(q[0])
        else:
            cnj = ' ' + conjunction + ' '
            return(cnj.join(q))


    def is_google_document(self):
        return(self.mime_type in googleapps.drive.MimeType.google_document_types)

    def is_media(self):
        return(self.mime_type in googleapps.drive.MimeType.media_types)

    def cache(self):
        # TODO
        # THIS IS NOT UPDATED


        path = self.cache_directory
        if not os.path.exists(path):
             os.makedirs(path)

        dt = datetime.datetime.now()
        folder_name = 'cache/' + dt.strftime('%Y-%m-%d %H.%M.%S')

        os.makedirs(folder_name)

        for network in self.get_networks():
            os.makedirs(folder_name + '/' + network.dirname)
            print("\nDownload and cache files for " + network.dirname)
            for plan in network.get_school_plans():
                fn = plan.filename
                if not re.search(r'\.xlsx$', fn, re.I):
                    fn += '.xlsx'
                self.download_file(plan.id,
                                   fn,
                                   format='xl',
                                   dir=folder_name + '/' + network.dirname)

        # if we get here, we are done!
        if os.path.exists('cache/current'):
            os.remove('cache/current')
        os.symlink(os.path.realpath(folder_name), 'cache/current', target_is_directory=True)
        # cache updated!

    def change_owner(self, new_owner):
        raise Exception("Not implemented yet.");

    def share(self, role, auth_type='user', share_with=None, domain=None,
            allow_discovery=None, expiration_date=None, with_link=None,
            message=None, notify=True, transfer_owner=False, domain_admin=False):
        """ Control sharing access to the folder.

        role {str} : The permission role: [owner, organizer, fileOrganizer, writer, commente, reader]
        type {str} : The type of grantee. One of: [user, group, domain, anyone]
        share_with {str|list}: The user email address(es) to share with.
        domain {str} (optional): The domain to which this new permission applies
        with_link {bool}: If true, set these permissions to anyone with the link. If with_link is true and domain is not None, it will apply to anyone with the link but in the domain
        """

        #---- General args, but the real action happens in the body (defined below)
        overall_args = {
            'fileId':self.id,
            'useDomainAdminAccess':domain_admin,
            'transferOwnership':transfer_owner,
        }

        # special case ??
        body = { 'role': role, 'type': auth_type }
#        if role == 'commenter':
#            body = { 'role': role, 'type': auth_type, 'additionalRoles': [ 'commenter' ] }
#        else:
#            body = { 'role': role, 'type': auth_type }

        if domain:
            body['domain'] = domain
        if expiration_date:
            # this should be in RFC 3339 date-time
            body['expirationDate'] = expiration_date
        if with_link is not None:
            body['withLink'] = with_link
        if allow_discovery is not None:
            body['allowFileDiscovery'] = allow_discovery

        if share_with:
            overall_args['sendNotificationEmail'] = notify
            overall_args['emailMessage'] = message

            if type(share_with) not in (list,tuple):
                share_with = [ share_with ]

            for email in share_with:
                single_body = body.copy()
                single_body['emailAddress'] = email
                self.add_request('permissions', 'create', body=single_body, **overall_args)
        else:
            self.add_request('permissions', 'create', body=body, **overall_args)



    def metadata(self, metadata=None, refresh=False):
        if self.deleted:
            raise FileNotFoundError("Attempt to use drive object after deletion")

        if self._metadata and not refresh:
            return(self._metadata)

        if not metadata:
            # https://developers.google.com/resources/api-libraries/documentation/drive/v3/python/latest/drive_v3.files.html#get
            try:
                metadata = self.default_service.files().get(fileId=self.id, fields='*').execute()
            except Exception as e:
                errmsg = e._get_reason()
                if re.search('File not found:', errmsg):
                    raise FileNotFoundError(errmsg)
                # anything else, raise like normal
                raise(e)

        self._metadata = metadata

        # -- create the current instance variable set based on the response.
        #    At the moment this is minimal. See the api-libraries for the full scope
        #    of information.

        # -- determine the type
        self._mime_type = metadata['mimeType']

        ## DETERIME
        # is_file
        # is_folder
        # ...
        self.type_isa = googleapps.drive.mime_type_isa(metadata['mimeType'])

        # Whether the file has been explicitly trashed, as opposed to recursively trashed from a parent folder.
        self._name = metadata["name"] # The name of the file. This is not necessarily unique within a folder. Note that for immutable items such as the top level folders of Team Drives, My Drive root folder, and Application Data folder the name is constant.
        if 'description' in metadata:
            self._description = metadata['description'] # A short description of the file.
        else:
            self._description = None
        # Whether the file has been trashed, either explicitly or from a trashed parent folder. Only the owner may trash a file, and other users cannot see files in the owner's trash.
        self._trashed = metadata['trashed']
        if 'size' in metadata:
            self._size = metadata['size']
        else:
            self._size = None
        if 'parents' in metadata:
            self._parent_folder_ids = metadata['parents']

        # this gets turned into an object on first call to obj.parents
        self._parents = None

        self._user_is_owner = metadata['ownedByMe']
        self._is_shared = metadata['shared']
        self._writers_can_share = metadata['writersCanShare']

        self._modified_user = metadata['lastModifyingUser']['permissionId']
        self._last_modified_user_email = metadata['lastModifyingUser']['emailAddress']
        #'is_me': metadata['lastModifyingUser']['me'],


        self._created_time = dateutil.parser.parse(metadata["createdTime"]) # The time at which the file was created (RFC 3339 date-time).
        self._modified_time = dateutil.parser.parse(metadata["modifiedTime"]) # The last time the file was modified by anyone (RFC 3339 date-time).

        self._ever_accessed_by_me = metadata["viewedByMe"] # Whether the file has been viewed by this user.
        self._ever_modified_by_me = metadata["modifiedByMe"] # Whether the file has been modified by this user.
        if "viewedByMeTime" in metadata:
            self._access_time_by_me = dateutil.parser.parse( metadata["viewedByMeTime"] )
        else:
            self._access_time_by_me = None

        if "modifiedByMeTime" in metadata:
            self._modified_time_by_me = dateutil.parser.parse( metadata["modifiedByMeTime"] )
        else:
            self._modified_time_by_me = None

        ###### DOES THIS WORK ON FILES AND NOT FOLDERS?
        if "folderColorRgb" in metadata:
            # only in folders
            self._folder_color = metadata["folderColorRgb"] # The color for a folder as an RGB hex string. The supported colors are published in the folderColorPalette field of the About resource.
            # If an unsupported color is specified, the closest color in the palette will be used instead.

        self._capabilities = {
         # Capabilities the current user has on this file. Each capability corresponds to a fine-grained action that a user may take.
          'can_read_revisions': metadata['capabilities']["canReadRevisions"], # Whether the current user can read the revisions resource of this file. For a Team Drive item, whether revisions of non-folder descendants of this item, or this item itself if it is not a folder, can be read.
          'can_trash': metadata['capabilities']["canTrash"], # Whether the current user can move this file to trash.
          'can_copy': metadata['capabilities']["canCopy"], # Whether the current user can copy this file. For a Team Drive item, whether the current user can copy non-folder descendants of this item, or this item itself if it is not a folder.
          'can_share': metadata['capabilities']["canShare"], # Whether the current user can modify the sharing settings for this file.
          'can_delete': metadata['capabilities']["canDelete"], # Whether the current user can delete this file.
          'can_rename': metadata['capabilities']["canRename"], # Whether the current user can rename this file.
          'can_list_children': metadata['capabilities']["canListChildren"], # Whether the current user can list the children of this folder. This is always false when the item is not a folder.
          'can_download': metadata['capabilities']["canDownload"], # Whether the current user can download this file.
          'can_comment': metadata['capabilities']["canComment"], # Whether the current user can comment on this file.
          'can_edit': metadata['capabilities']["canEdit"], # Whether the current user can edit this file.
          'can_remove_children': metadata['capabilities']["canRemoveChildren"], # Whether the current user can remove children from this folder. This is always false when the item is not a folder.
          'can_add_children': metadata['capabilities']["canAddChildren"], # Whether the current user can add children to this folder. This is always false when the item is not a folder.
          'can_change_viewers_can_copy_content': metadata['capabilities']["canChangeViewersCanCopyContent"], # Whether the current user can change whether viewers can copy the contents of this file.
          'can_untrash': metadata['capabilities']["canUntrash"], # Whether the current user can restore this file from trash.

            # --- not enabled because of teamdrive reference (and may generate keyerror
#          'can_move_into_teamdrive': metadata['capabilities']["canMoveItemIntoTeamDrive"], # Whether the current user can move this item into a Team Drive. If the item is in a Team Drive, this field is equivalent to canMoveTeamDriveItem.
#          'can_move_teamdrive': metadata['capabilities']["canMoveTeamDriveItem"], # Whether the current user can move this Team Drive item by changing its parent. Note that a request to change the parent for this item may still fail depending on the new parent that is being added. Only populated for Team Drive files.
            # 'can_read_teamdrvie': metadata['capabilities']["canReadTeamDrive"], # Whether the current user can read the Team Drive to which this file belongs. Only populated for Team Drive files.

        }


        # The owners of the file. Currently, only certain legacy files may have more than one owner. Not populated for Team Drive files.
        if len(metadata['owners']):
            self._owners = [
                {'is_me' : owner["me"], # Whether this user is the requesting user.
                'user_display_name': owner["displayName"],  # A plain text displayable name for this user.
                'user_id': owner["permissionId"], # The user's ID as visible in Permission resources.
                'user_email': owner["emailAddress"]} # The email address of the user. This may not be present in certain contexts if the user has not made their email address visible to the requester.
                for owner in metadata['owners']]
        else:
            self._owner = None

        if 'md5Checksum' in metadata:
            self._md5 = metadata["md5Checksum"] # The MD5 checksum for the content of the file. This is only applicable to files with binary content in Drive.
        else:
            self._md5 = None

        if "webContentLink" in metadata:
            self._download_link = metadata["webContentLink"] # A link for downloading the content of the file in a browser. This is only available for files with binary content in Drive.
        else:
            self._download_link = None
        self._users_with_access = metadata["permissionIds"]

        # The final component of fullFileExtension. This is only available for files with binary content in Drive.
        # The full file extension extracted from the name field. May contain multiple concatenated extensions, such as "tar.gz". This is only available for files with binary content in Drive.
        if 'fullFileExtension' in metadata:
            self._full_file_extension = metadata['fullFileExtension']
        else:
            self._full_file_extension = None
        # This is automatically updated when the name field changes, however it is not cleared if the new name does not contain a valid extension.
        if 'fileExtension' in metadata:
            self._file_extension = metadata["fileExtension"]
        else:
            self._full_file_extension = None

        return(self._metadata)



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import googleapps
from .driveitem import DriveItem
from .file import File
from .folder import Folder

import googleapiclient

import os, os.path
import datetime
import pickle
import re
import shutil

import warnings
from pprint import pprint


file_type_mimetype = {
    'file': 'application/vnd.google-apps.file',
    'folder': 'application/vnd.google-apps.folder',

    'document': 'application/vnd.google-apps.document',
    'drawing': 'application/vnd.google-apps.drawing',
    'form': 'application/vnd.google-apps.form',
    'presentation': 'application/vnd.google-apps.presentation',
    'spreadsheet': 'application/vnd.google-apps.spreadsheet',

    'audio': 'application/vnd.google-apps.audio',
    'video': 'application/vnd.google-apps.video',
    'photo': 'application/vnd.google-apps.photo',
    
    'map': 'application/vnd.google-apps.map',
    'script': 'application/vnd.google-apps.script',
    'site': 'application/vnd.google-apps.site',
    'fusiontable': 'application/vnd.google-apps.fusiontable',

    #'application/vnd.google-apps.drive-sdk': ['file', ],
    #    'application/vnd.google-apps.unknown': ['file', ],


    'zip':'application/zip',
    
    'pdf':'application/pdf',
    'epub':'application/epub+zip',
    'svg':'image/svg+xml',
    'png':'image/png',
    
    'html':'text/html',
    'txt':'text/plain',
    'text':'text/plain',
    'csv':'text/csv',
    'tsv':'text/tab-separated-values',

    'rtf':'application/rtf',
    
    'docx':'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    'xlsx':'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    'pptx':'application/vnd.openxmlformats-officedocument.presentationml.presentation',

    'odt':'application/vnd.oasis.opendocument.text',
    'ods':'application/x-vnd.oasis.opendocument.spreadsheet',
    'odg':'image/jpeg',
    'odp':'application/vnd.oasis.opendocument.presentation',
    'script':'application/vnd.google-apps.script+json',
}

doctype_search = (
    [('doc','docx','odt','rtf'), 'document'],
    [('xls','xlsx','ods','csv','tsv'),'spreadsheet'],
    [('pptx','ppt','odp'), 'presentation'],
    [('odg'),'drawing'],
)

def is_document(extension):
    for doctype in doctype_search:
        if extension in doctype[0]:
            return(doctype[1])
    return

def translate_filetype(file_type):
    if file_type in file_type_mimetype:
        return(file_type_mimetype[file_type])
    return

_mimetype_isa = {}
def mimetype_isa(mime_type):
    if not _mimetype_isa:
        for drivetype, mimetype in file_type_mimetype.items():
            if drivetype in ('file', 'folder'):
                _mimetype_isa[mimetype] = [drivetype]        
            else:
                _mimetype_isa[mimetype] = [drivetype, 'file']        

    if mime_type in _mimetype_isa:
        return(_mimetype_isa[mime_type])
    # default is just file
    return(['file'])


def create(name, in_folder=None, path=None, **body):

    body['name'] = name

    if in_folder:
        if type(in_folder) == 'str':
            body['parents'] = [ in_folder ]
        else:
            body['parents'] = [ in_folder.id ]

    if path:
        raise Exception("Not implemented yet")

    # create the directory
    cls = DriveItem
    response = cls.services.load_service('drive').files().create(
        body=body,
        fields='*',
    ).execute()

    #-- create the self object
    # -- now we reclassify the type because DriveItem is kinda an abstract class
    subclass = cls.translate_mimetype(response['mimeType'])
    if subclass:
        self = subclass(response['id'], metadata=response)
    else:
        # default: just treat it like a file
        self = File(response['id'], metadata=response)

    # return!
    return(self)


def upload(filepath, source_mime_type=None, name=None, in_folder=None, path=None, mime_type=None, convert=True, resumable=True, **parameters):

    metadata = {}
    for field, value in parameters.items():
        metadata[ googleapps.toCamel(field) ] = value

    # get the filename
    match = re.search(r'([^\/\\]+)$', filepath)
    filename = match.group(1)

    if name:
        metadata['name'] = name
    else:
        metadata['name'] = filename

    if in_folder:
        if type(in_folder) == 'str':
            metadata['parents'] = [ in_folder ]
        else:
            metadata['parents'] = [ in_folder.id ]

    if path:
        raise Exception("Not implemented yet")

    match = re.search(r'\.([^\.]+)$', filename)
    if match:
        extension = match.group(1).lower()

    #now get/load the source file
    if not os.path.exists(filepath):
        raise FileNotFoundError("Source file to upload does not exist at path " + str(filepath))

    media = googleapiclient.http.MediaFileUpload(filepath, mimetype=source_mime_type, resumable=resumable)

    if convert:
        # turn it into a google document (if it is an elgible type)            
        doctype = is_document(extension)
        if doctype:
            metadata['mimeType'] = translate_filetype(doctype)

    cls = DriveItem
    response = cls.services.load_service('drive').files().create(
        body=metadata,
        media_body=media,
        fields='*',
    ).execute()

    subclass = cls.translate_mimetype(response['mimeType'])
    if subclass:
        self = subclass(response['id'], metadata=response)
    else:
        self = File(response['id'], metadata=response)
    return(self)



def find_item(*args, **kwargs):
    # shortcut to find exactly one item
    result = find(*args, **kwargs)
    if not result:
        return
    if len(result) == 1:
        return result[0]
    raise Exception("Parameters sent to find_file() matched multiple files: " + "; ".join(f.name + '( ' + f.id + ' )' for f in result))



def find(filename=None, file_type=None, in_folder=None, domains=None, spaces=None, order=None, **queryfields):

    params = {}
    if domains:
        params['corpora'] = domains

    if order:
        if 'order_by' in queryfields:
            raise Exception('Cannot specify both order_by and order')
        params['order_by'] = order

    if spaces:
        params['spaces'] = spaces

    if filename:
        if 'name' in queryfields:
            raise Exception("Cannot specify both name and filename. These are the same")
        queryfields['name'] = filename

    if file_type:
        if 'mime_type' in queryfields:
            raise Exception('Cannot specify both file_type and mime_type')
        queryfields['mime_type'] = translate_filetype(file_type)
        
    if in_folder:
        if 'parents' in queryfields:
            raise Exception("Cannot specify in_folder and parents")
        if type(in_folder) is str:
            queryfields['parents'] = {'op': 'in', 'value': in_folder}
        else:
            queryfields['parents'] = {'op': 'in', 'value': in_folder.id}

    # by default, we do not serch in the trash
    if 'trashed' not in queryfields:
        queryfields['trashed'] = False


    cls = DriveItem
    q = cls.generate_query_string(**queryfields)

    params['q'] = q
    print("QUERY is " + q)
    response = cls.services.load_service('drive').files().list(
        fields='*',
        **params
    ).execute()

    if not len(response['files']):
        return

    result = []
    for f in response['files']:
        result.append(cls(f['id'], metadata=f))
    return(result)






# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import googleapps
class ListElement(googleapps.JsonBase):

    def __init__(self, list_id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = list_id
        self._levels = None

    @property
    def levels(self):
        if not self._levels:
            self._levels = [ListLevel(j, self) for j in self._json['listProperties']['nestingLevels']]
        return(self._levels)

    def get_level(self, level):
        return(self.levels[level])

class ListLevel(googleapps.JsonBase):

    @property
    def list_type(self):
        if self.glyph_symbol:
            return('unordered')
        elif self.glyph_type == 'NONE':
            return('none')
        else:
            return('ordered')

    @property
    def item_type(self):
        list_type = self.list_type
        if list_type == 'unordered':
            return(self.glyph_symbol)
        elif list_type == 'none':
            return(None)

        ordered_type = self.glyph_type
        if ordered_type in ('DECIMAL', 'ZERO_DECIMAL'):
            return('numeric')
        if ordered_type in ('ALPHA', 'UPPER_ALPHA'):
            return('alpha')
        if ordered_type in ('ROMAN', 'UPPER_ROMAN'):
            return('roman')
        raise Exception(f"Unkonwn ordered type {ordered_type}")

    @property
    def glyph_type(self):
        if 'glyphType' in self.json:
            return(self.json['glyphType'])
        return(None)

    @property
    def glyph_symbol(self):
        if 'glyphSymbol' in self.json:
            return(self.json['glyphSymbol'])
        return(None)



import re
import googleapps
from .spreadsheet import Spreadsheet
from .worksheet import Worksheet
from .format import Format
#import googleapps.sheets.xlsxwriter
#import googleapps.sheets.openpyxl


#------------- UTILITY FUNCTIONS -----------------------

ord_a = ord('A')
def column_letter_to_integer(col):
    col = col.upper()
    strlen = len(col)
    column = 0
    for offset in range(strlen):
        column += (ord(col[strlen-offset-1]) - ord_a + 1) * 26**offset
    column -= 1
    return(column)

def column_offset_to_letter(column):
    # this translates a 0-indexed column to the appropriate letter
    if column > 25:
        string = chr(ord_a + column % 26)
        column = int(column / 26)
        while column:
            string = chr(ord_a + (column-1) % 26) + string
            column = int((column-1) / 26)
        # this is the first digit, which needs 1 subtracted for it
        # because of the letter - number conversion issue
        return(string)
    return(chr(ord_a+column))


# Presently, we do allow 'a1' if people are lazy.
# Should it be constrained to only be uppercase? A1? AA1
a1re = re.compile(r'([A-Za-z]+)(\d+)$')
def a1_to_rowcol(a1cell):
    m = a1re.match(a1cell)
    if not m:
        raise Exception("'"+a1cell+"' does not appear to actually be in A1 notation. Cannot convert.")
    return(int(m.group(2)) - 1, column_letter_to_integer(m.group(1)))

a1rangere = re.compile(r'([A-Za-z]+)(\d+):([A-Za-z]+)(\d+)$')
def parse_a1_range(a1range):
    m = a1rangere.match(a1range)
    if not m:
        # to be resilient, we'll allow "A1" to be considered a range from [1,0,1,0]
        m = a1re.match(a1range)
        if not m:
            raise Exception("'" + a1range + "' does not appear to actually be in an A1:B2 notation. Cannot parse/convert.")
        return(int(m.group(2)) - 1, column_letter_to_integer(m.group(1)), int(m.group(2)) - 1, column_letter_to_integer(m.group(1)))
    return(int(m.group(2)) - 1, column_letter_to_integer(m.group(1)), int(m.group(4)) - 1, column_letter_to_integer(m.group(3)))

colspanre = re.compile(r'([A-Za-z]+):([A-Za-z]+)$')
def parse_colspan(colspan):
    if len(colspan) == 1:
        diff = ord(colspan) - ord_a
        return(diff, diff)
    m = colspanre.match(colspan)
    return (column_letter_to_integer(m.group(1)), column_letter_to_integer(m.group(2)))

def range_to_a1(row, column, last_row=None, last_column=None, worksheet=None):
    if worksheet:
        if type(worksheet) is str:
            result = worksheet + '!'
        else:
            result = worksheet.title + '!'
    else:
        result = ""

    result += chr(ord_a + column) + str(row+1)
    if last_row:
        result += ':' + column_offset_to_letter(last_column) + str(last_row+1)

    return(result)

def create(title, developer_metadata=None, worksheet_data=None, named_ranges=None, spreadsheet_params=None):
    # https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#SpreadsheetProperties
    if spreadsheet_params:
        spreadsheet_properties = spreadsheet_params.copy()
    else:
        spreadsheet_properties = {}

    spreadsheet_properties['title'] = title
    # now the actual create call
    self = Spreadsheet()
    self.add_request('spreadsheets', 'create',
        callback=self.process_json,
        body={
            'properties': spreadsheet_properties,
            'sheets': worksheet_data,
            'namedRanges': named_ranges,
            'developerMetadata': developer_metadata,
        })

    #---- At present, the first sheet always has index 0 and id 0
    #     So, we can create it in place without actually commiting the callback
    self._add_uncommited_worksheet(title='Sheet1', id=0, index=0)

    return(self)

urlre = re.compile(r'https://(www\.)?(docs.google.com/spreadsheets/d/([\w\-]+))?', re.I)
def get(id, ranges=None, include_grid=False):
    #-- we will manually process full urls instead of straight ids
    isurl = urlre.match(id)
    if isurl:
        actual_id = isurl.group(3)
        if not actual_id:
            raise ValueError("Call to googleapps.sheets.get provided a URL but not an expected url to a Google spreadsheet: " + id)
        id = actual_id

    self = Spreadsheet()
    self.make_request('spreadsheets', 'get',
        callback=self.process_json,
        spreadsheetId=id, ranges=ranges, includeGridData=include_grid
    )
    return(self)



# import classes into the googleapps namespace
from googleapps.service import ServiceManager
from googleapps.serviceclient import ServiceClient
from googleapps.base import *

# import core api modules
import googleapps.sheets
import googleapps.drive

import re
import datetime, time

# requires apiclient, oauth2client, google-api-python-client

""" googleapps

This module provides a suite of encapsulations of API access to Google Apps that attempt to follow encapsulation expectations for good object-oriented (and Python) code, as much as possible. So, instead of the user needing to parse Google API references (https://developers.google.com/apis-explorer/) or rely on individual wrappers that encapsulate single APIs, this library seeks to:
1. Provide a consistent set of base packages to access the API and generate low-level but object-oriented codebooks
2. Provide a system to build well encapsulated systems upon it
3. Implement core APIs that I use into this sytem (ie. drive, sheets, and gmail at the time of this documentation)

Also, because rapid development (for me) often starts off by creating local versions that I realize are better off in the Google ecosystem, I follow the following Guidelines in module development:
- Use naming conventions and functional encapsulation that parallels common local-source alternatives so that converting between one system and another (or maintaining local and remote options) requires as little work for the coder as possible.
- In line with #1, if the Google API is conceptually different, follow local module naming conventions for the core googleapps module and provide an additional function that follows the Google API encapsulation, so that developers familiar with the Google reference can work seemlessly
"""

parens = re.compile(r'_(\w)')
humps = re.compile(r'([a-z])([A-Z0-9]+)')
numletter = re.compile(r'(\d)([a-zA-Z]+)')
acronym = re.compile(r'([A-Z]+)([A-Z][a-z\d])')

def toCamel(string:str) -> str:
    """ convert snake_case into camelCase. (Note: this is not purely reciprocal with fromCamel because there is a loss of resolution of uppercase acronyms in the conversion to snake case)"""
    return(parens.sub(lambda m: m.group(1).upper(), string.strip()))

def fromCamel(string:str) -> str:
    """ convert camelCase into snake_case """
    string = humps.sub(lambda m: m.group(1) + '_' + m.group(2), string.strip())
    string = numletter.sub(lambda m: m.group(1) + '_' + m.group(2), string)
    string = acronym.sub(lambda m: m.group(1) + '_' + m.group(2), string)
    return(string.lower())

def unCamel(json:dict, flatten=False) -> dict:
    """ Take an arbitrary data structure with camelCase keys and turn it into a snake_case python data structure. The most frequent use case is when receiving json from Google API calls.
    TODO: Resolve whether integer-looking keys should be remapped to actual integers (since json requires all keys to be strings. I'm not convinced this is actually ever a priority in the Google API case, but it is a major difference between the JSON spec and Python datastructures in general - so, be careful using this for purposes aside from parsing REST responses)

    Args:
        json (object): The data structure to unCamel.
        flatten (bool, optional): If true, do not not duplicate the nesting part of nested dictionaries, instead ignore the top-level key and move all bottom-level keys into the result tree.  Default is false.

    Examples:
        >>> print unCamel({'camelParamA': 12, 'camelParamB': [1,2,3], 'camelParamC': {'subCamelA': 1, 'subCamelB': 2}})
        {'camel_param_a': 12, 'camel_param_b': [1, 2, 3], 'camel_param_c': {'sub_camel_a': 1, 'sub_camel_b': 2}}

        >>> print unCamel({'camelParamA': 12, 'camelParamB': [1,2,3], 'camelParamC': {'subCamelA': 1, 'subCamelB': 2}}, flatten=True)
        {'camel_param_a': 12, 'camel_param_b': [1, 2, 3], 'sub_camel_a': 1, 'sub_camel_b': 2} # no 'camel_param_c'!!
    """
    if not json:
        return
    # handle non-dict types
    if type(json) in (list, tuple):
        return(list(map(lambda item: unCamel(item), json)))
    elif type(json) is not dict:
        return(json)

    # okay, expected json struct
    result = {}
    for key in json:
        if type(json[key]) is dict:
            subjson = unCamel(json[key], flatten)
            if flatten:
                result.update(subjson)
            else:
                result[fromCamel(key)] = subjson
        elif type(json[key]) in (list, tuple):
            # lists are possibly atomic
            # but they could also be json substructs
            # we figure that out by looking at the first element
            # if it's a dict, we process
            # else we don't
            new_key = fromCamel(key)
            if not len(json[key]) or type(json[key][0]) is not dict:
                result[new_key] = json[key]
            else:
                processed = list(map(lambda element: unCamel(element), json[key]))
                if flatten:
                    # this is a major problem, right?
                    raise Exception("This probalby shouldn't be allowed as it would likely overwrite itself")
                    for subjson in processed:
                        result.update(subjson)
                else:
                    result[new_key] = processed
        else:
            result[fromCamel(key)] = json[key]
    return(result)

def to_rfctime(datetimeobj) -> str:
    """ Converts a datetime.* object or unix epoch time to an RFC 3339 time string.

    Args:
        datetimeobj: One of the following (note that wihtout an explicit timezone, localtime is used):
            - a datetime.date or datetime.datetime object, which is directly converted (dates will return 00:00:00 as their time)
            - a datetime.time object, which is converted to using today at the given time. Use with caution.
            - an integer of float that represents unix epoch (local)time, the number of seconds since 1970-01-01 in the local timezone

    Returns: A string of the argument in RFC3339 format
    """
    t = type(datetimeobj)
    # -- handle epoch time and datetime.date recursively
    if t in (int, float):
        return(to_rfctime(datetime.datetime.fromtimestamp(datetimeobj)))
    elif t is datetime.date:
        return(to_rfctime(datetime.datetime(datetimeobj.year, datetimeobj.month, datetimeobj.day)))

    tzoffset = datetimeobj.utcoffset()

    def tzstring_from_seconds(seconds):
        difftime = seconds / 3600
        hours = int(difftime)
        if hours < 0:
            return(format(hours, '03d') + ':' + format(int((hours-difftime)*60), '02d'))
        else:
            return('+' + format(hours, '02d') + ':' + format(int((difftime-hours)*60), '02d'))

    if tzoffset is not None:
        secoffset = tzoffset.total_seconds()
        if secoffset == 0:
            tzstring = 'Z'
        else:
            # NOTE that secoffset from the datetime object will be negative for western hemisphere timezones!!
            tzstring = tzstring_from_seconds(secoffset)
    else:
        # set to LOCALTIME
        # NOTE THAT LOCALTIME offsets will be POSITIVE for western hemisphere timezones!!
        lt = time.localtime(datetimeobj.timestamp())
        # check for daylight savings time
        if lt.tm_isdst:
            tzstring = tzstring_from_seconds(-time.altzone)
        else:
            tzstring = tzstring_from_seconds(-time.timezone)

    if datetimeobj.microsecond:
        # we dont want a string that is T12:15:01.420000, so we reformat the microsecond argument
        ms = str(datetimeobj.microsecond / 1000000)[1:]
    else:
        ms = ''
    if t is datetime.datetime:
        return(datetimeobj.strftime("%Y-%m-%dT%H:%M:%S") + ms + tzstring)
    elif t is datetime.time:
        today = datetime.date.now()
        return(today.strftime("%Y-%m-%dT") + datetimeobj.strftime("%H:%M:%SZ"))
    elif t in (int, float):
        # assume it's an epoch timestamp
        return(to_rfctime(datetime.datetime.fromtimestamp(datetimeobj)))
    return        

rfcre = re.compile('([0-9]{4})-(0[1-9]|1[012])-(0[1-9]|[12][0-9]|3[01])[Tt]([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.[0-9]+)?(([Zz])|(([\+|\-][01][0-9]|2[0-3]):([0-5][0-9])))')
def from_rfctime(rfcstring, require=False) -> datetime.datetime:
    """ Converts an RFC time string to a datetime.datetime object.

    Args:
        rfcstring (str): The string to convert into a datetime object. Note that if require is False, the type is not checked.
        require (bool): If true, guarantees that the function returns a datetime object or throws an exception. This allows the user to avoid third-party input verification.
    Return:
        None if rfcstring is not a string in rfc format
    """
    if not type(rfcstring) is str:
        if require:
            raise ValueError("Input to from_rfctime is not a string, it is " + str(type(rfcstring)) )
        return
    m = rfcre.fullmatch(rfcstring)
    if not m:
        if require:
            raise ValueError("Input '"+rfcstring+"' to from_rfctime is not in RFC3334 format" )
        return

    parts = m.groups()
    # convert the regex match to integers
    params = [int(x) for x in parts[:6]]

    # resolve mseconds
    if parts[6] is not None:
        params.append(int(float(parts[6])*1000000))

    if parts[7] == 'Z':
        # utc timezone
        return(datetime.datetime(*params, tzinfo=datetime.timezone.utc))
    else:
        # convert to local timezone
        hours = int(parts[10])
        minutes = int(parts[11])
        return(datetime.datetime(*params, tzinfo=datetime.timezone(datetime.timedelta(hours=hours, minutes=minutes))))

    # old logic
    #match = rfcre.match(rfcstring)
    #if not match:
    #    raise Exception("Expected string in RFC format, but this did not match the regex: " + str(rfcstring))
    #rfc = match.groups()
    #return(datetime.datetime(*[int(x) for x in [*rfc[0:5], rfc[6]]]))

def json_fieldmask(json):
    """ Converts a dict with nested dict objects into a list in which nested keys are appended to the parent key with a dot. The general use case is when the Google API requests a list of fields and the user wishes to autogenerate the fields as exactly the parameters set in a python object that is going to be sent.

    Example:
        >>>json_fieldmask({'ParamA': 12, 'ParamB': [1,2,3], 'ParamC': {'subA': 1, 'subB': 2}})
        ['ParamA',
         'ParamB',
         'ParamC.subA',
         'ParamC.subB']
        """
    fields = []
    for field in json:
        if type(json[field]) is dict:
            subfields = json_fieldmask(json[field])
            fields.extend([ '.'.join([field, sf]) for sf in subfields ])
        else:
            fields.append(field)
    return(fields)



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import googleapps
import datetime

@googleapps.jsonbase
class EventTime(googleapps.JsonBase):
    json_properties = ['date', 'date_time', 'time_zone']

@googleapps.jsonbase
class Event(googleapps.JsonBase):
    json_properties = ['id', 'status', 'html_link', 'created', 'updated', 'summary', 'description', 'location', 'creator', 'organizer', 'start', 'end', 'end_time_unspecified', 'recurrence', 'attendees']

    json_objects = {
        'start': EventTime,
        'end': EventTime,
        'originalStartTime': EventTime,
    }

    @classmethod
    def create(cls, service, calendar_id, summary, start, end, **kwargs):
        if type(start) is datetime.date:
            start = {
                'date': start.isoformat(),
            }
            end = {
                'date': end.isoformat(),
            }
        else:
            start = {
                'date_time': googleapps.to_rfctime(start),
                'time_zone': 'America/New_York'
            }
            end = {
                'date_time': googleapps.to_rfctime(end),
                'time_zone': 'America/New_York'
            }

        eventjson = service.make_request('events', 'insert', calendarId=calendar_id,
            body={
                'start':start,
                'end':end,
                'summary': summary,
                **kwargs,
        })
        return(Event(eventjson))


@googleapps.jsonbase
class EventList(googleapps.JsonBase):
    json_properties = ['etag', 'summary', 'description', 'updated', 'time_zone', 'access_role', 'default_reminders']
    json_objects = {
        'items': Event
    }

    def __getitem__(self, i):
        """ This meta-method allows the EventList object to be treated like an indexed array of Events.

        Example:
            import usl
            u = usl.USL()
            school_0 = u[0]
            """
        return(self.items[i])

    def __len__(self):
        """ This meta-method allows len() to work correctly """
        return(len(self.items))




import re
import warnings
from pprint import pprint

import googleapps.sheets

POINT_HEIGHT_TO_INCH = 72
POINT_WIDTH_TO_INCH = 12
PIXEL_TO_INCH = 96

POINT_HEIGHT_TO_PIXEL = PIXEL_TO_INCH / POINT_HEIGHT_TO_INCH
POINT_WIDTH_TO_PIXEL = PIXEL_TO_INCH / POINT_WIDTH_TO_INCH

class Worksheet(googleapps.sheets.Worksheet):

    @googleapps.sheets.worksheet.rowcol_method
    def write(self, row, col, value, format=None):
        return(self.update_cells(row, col, value, format))
    
    @googleapps.sheets.worksheet.rowcol_method
    def write_row(self, row, col, data, format=None):
        self.update_cells(row, col, data, format)

    @googleapps.sheets.worksheet.rowcol_method
    def write_rows(self, row, col, dataset, format=None):
        self.update_cells(row, col, dataset, format)

    #---- All of these write methods could just alias write() above, but then they wouldn't 
    #     provide any typechecking
    
    @googleapps.sheets.worksheet.rowcol_method
    def write_formula(self, row, col, value, format=None):
        if type(value) is not str or value[0] != '=':
            raise Exception("Value sent to write_formula not a formula: " + str(value))
        self.update_cells(row, col, {'userEnteredValue': {'formulaValue': value}}, format)

    @googleapps.sheets.worksheet.rowcol_method
    def write_number(self, row, col, value, format=None):
        self.update_cells(row, col, {'userEnteredValue': {'numberValue': value}}, format)

    @googleapps.sheets.worksheet.rowcol_method
    def write_boolean(self, row, col, value, format=None):
        self.update_cells(row, col, {'userEnteredValue': {'booleanValue': value}}, format)

    @googleapps.sheets.worksheet.rowcol_method
    def write_url(self, row, col, url, format, string=None, tip=None):
        raise Exception("write_url not implemented yet")

    @googleapps.sheets.worksheet.rowcol_method
    def write_datetime(self, row, col, value, date_format=None):
        raise Exception("this requires some additional thought and testing")
        self._update_cells(row, col, {'numberValue': {'formulaValue': value}}, format)

    
    @googleapps.sheets.worksheet.sheetrange_method
    def merge_range(self, start_row, start_col, last_row, last_col, data=None, format=None):
        
        self.merge_cells(start_row, start_col, last_row, last_col)
        # Based on prior comments here, there can be issues here if the format includes a locked parameter
        # because the entire range has to be protected/unprotected, or the google client will throw an error

        #if 'locked' in format:
        if format.locked is not None:
            if format.locked:
                self.protect(start_row, start_col, last_row, last_col)
            else:
                self.unprotect(start_row, start_col, last_row, last_col)
            #format = format.copy()
            #del format['locked']
            
        # okay, now update the cells with the data and any additional format
        return(self.update_cells(start_row, start_col, data, format))
    
    
    def set_column(self, colspan, width):
        start,end = googleapps.sheets.parse_colspan(colspan)
        self.update_columns(start,end, size=width*POINT_WIDTH_TO_PIXEL)
        
    def set_row(self, row, height):
        self.update_rows(row,row, size=int(height*POINT_HEIGHT_TO_PIXEL))
        
    @googleapps.sheets.worksheet.sheetrange_method
    def data_validation(self, first_row, first_col, last_row, last_col, options):
        # TODO: This is only partially/minimally implemented
        val_type = options['validate']
        
        condition_params = {}
        if 'source' in options:
            condition_params['values'] = options['source']
            
        if 'error_type' in options:
            condition_params['required'] = options['error_type'].lower() == 'stop'
        
        if 'input_message' in options:
            condition_params['input_message'] = options['input_message']
#        if '' in options:
#            condition_params[''] = options['']
#        if '' in options:
#            condition_params[''] = options['']
#        if '' in options:
#            condition_params[''] = options['']
        
        
        if val_type == 'list':
            return(self.add_data_validation(
                first_row, first_col, last_row, last_col, 
                'ONE_OF_LIST', show_dropdown=True, **condition_params) 
            )
        else:
            raise Exception("Not implmented data validation of type " + str(val_type))
        
    @googleapps.sheets.worksheet.sheetrange_method
    def conditional_format(self, first_row, first_col, last_row, last_col, options):
        extra = {}
        if 'value' in options:
            extra['values'] = [options['value']] 
        elif 'values' in options:
            extra['values'] = options['values']
        
        self.add_conditional_formatting(
            first_row, first_col, last_row, last_col,
            options['criteria'], 
            format=options['format'],
            **extra             
        )
                
    @googleapps.sheets.worksheet.rowcol_method
    def freeze_panes(self, row, col, top_row=None, top_col=None):
        if top_row or top_col:
           warnings.warn('Note: Google sheets does not seem to support top_row/top_col')
        if not row:
            super().freeze_columns(col)
        elif not col:
            super().freeze_rows(row)
        else:
            super().freeze_panes(row, col)
        
    @googleapps.sheets.worksheet.rowcol_method
    def update_cells(self, row, column, values=None, format=None):
        result = super().update_cells(row, column, values, format)
        
        # the only thing extra we do is look to see if the format is locked,
        # which requires an additional call to proect the call      
        if format and format.locked is not None:
            if type(values) is list:
                if type(values[0]) is list:
                    last_row = row + len(values) - 1
                    last_column= column + len(values[0]) - 1
                else:
                    last_row = row
                    last_column= column + len(values) - 1
            else:
                last_row = row
                last_column = column
            if format.locked:
                self.protect(row, column, last_row, last_column)
            else:
                self.unprotect(row, column, last_row, last_column)
        return(result)
    
    def _process_format(self, ref):
        if type(ref) is dict:
            return 
        else:
            return ref.cellformat_json()




# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import re
import googleapiclient.errors

import googleapps
from googleapps.serviceclient import CommitType
import copy
import datetime
import warnings
from pprint import pprint


"""
rows_available, columns_available: the boundary of spreadsheet: the number of rows or columns, regardless of whether any data is in them
nrow, ncol: the last row and column with any data in it.  Calling this requires a full data fetch. Calling last column furthermore requires iterating through every row.


"""

serial_number_base = datetime.datetime(1899,12,30)

#-------- Class-based decoratior functions
def rowcol_method(func):
    def wrapper(self, *args, **kwargs):
        if len(args) and type(args[0]) is str:
            # we expect this to be in A1:C3 format
            args = googleapps.sheets.a1_to_rowcol(args[0]) + list(args[1:])
        return(func(self, *args, **kwargs))
    return(wrapper)

def sheetrange_method(func):
    def wrapper(self, *args, **kwargs):
        if len(args) and type(args[0]) is str:
            # we expect this to be in A1:C3 format
            args = list(googleapps.sheets.parse_a1_range(args[0])) + list(args[1:])
        return(func(self, *args, **kwargs))
    return(wrapper)

def colspan_method(func):
    def wrapper(self, *args, **kwargs):
        if len(args) and type(args[0]) is str:
            # we expect this to be in A1:C3 format
            args = googleapps.sheets.parse_colspan(args[0]) + list(args[1:])
        return(func(self, *args, **kwargs))
    return(wrapper)




@googleapps.serviceclient.googleappclient
class Worksheet(googleapps.ServiceClient):
    default_api = 'sheets'
    default_api_version = 'v4'
    commit_policy=CommitType.defer_commit
    colors = {
        'black': {'red':0, 'green':0, 'blue':0},
        'white': {'red':1, 'green':1, 'blue':1},
        'blue': {'red':0, 'green':0, 'blue':1},
        'red': {'red':1, 'green':0, 'blue':0},
        'green': {'red':0, 'green':1, 'blue':0},
        'purple': {'red':1, 'green':0, 'blue':1},
        }

    api_ro_properties = ('id',)
    api_properties = ('title', 'index', 'sheet_type', 'hidden', 'right_to_left', 'tab_color',
                        'rows_available', 'columns_available', 'frozen_rows_available', 'frozen_column_count', 'hide_gridlines')


    def clear(self):
        if self.nrows:
            # just in case ncol has never been calculated, becuase it is an expensive procedure,
            # we're going to use columns_available
            self.update_cells(0,0,[list(map(lambda x: None, range(self.columns_available))) for row in range(self.nrows)])
            self._celldata=None
        # if last row is 0, it's already clear

    def delete(self):
        #TODO: handle delete if this is not actually a committed sheet
        #self.spreadsheet.batch_update( 'deleteSheet', sheetId=lambda s: s.id )
        self.spreadsheet.batch_update( 'deleteSheet', sheetId=self.id )
        for i in range(len(self.spreadsheet._worksheets)):
            if self.spreadsheet._worksheets[i].id == self.id:
                self.spreadsheet._worksheets.pop(i)
                break

    def copy_to(self, spreadsheet, copy_protection=True, title=None, keep_title=False):
        """ Takes an ID or a googleapps.sheets.Spreadsheet object, copies this worksheet into it. This is done in real time, not deferred

        Args:
            spreadsheet (googleapps.spreadsheet|str): An object or ID for the spreadsheet to which this worksheet will be copied.
            copy_protection (bool): If this worksheet has protected ranges, should they be copied? The underlying API does not copy protected data. If the owner of the destination spreadsheet is not the same as the origin spreadsheet, there may be some differences in who has edit access as the owner can always edit the spreadsheet. Defaults to True.
            title (str, optional): Specify the title of the new worksheet.
            keep_title (bool): Indicates that the title of the new worksheet shoudl be the same as the source. By default, the underlying google API will call the spreadsheet "Copy of [original title]", possibly with an iterated number after it. This will replace that name with the original title of the source.  This should only be used if the destination spreadsheet is not the source as it will immediately create a duplicate name error. Default to False.
        """

        self.commit()
        if type(spreadsheet) is str:
            spreadsheet_id = spreadsheet
            spreadsheet = None
        else:
            spreadsheet_id = spreadsheet.id
        req = self.make_request(['spreadsheets', 'sheets'],
                          'copyTo',
                          spreadsheetId=self.spreadsheet.id,
                          sheetId=self.id,
                          body={'destination_spreadsheet_id': spreadsheet_id},
        )
        if not spreadsheet:
            spreadsheet = googleapps.sheets.get(spreadsheet_id)
        new_ws = Worksheet(spreadsheet, json={'properties':req})
        if self.protected_ranges and copy_protection:
            # for next stage debuging
            for protected_range in self.protected_ranges:
                copy_protection = {}
                if 'description' in protected_range:
                    copy_protection['description'] = protected_range['description']
                if 'warning_only' in protected_range:
                    copy_protection['warning_only'] = protected_range['warning_only']
                if 'editors' in protected_range:
                    if 'users' in protected_range['editors']:
                        copy_protection['user_editors'] = protected_range['editors']['users']
                    if 'groups' in protected_range['editors']:
                        copy_protection['group_editors'] = protected_range['editors']['groups']
                    if 'domain_users_can_edit' in protected_range['editors']:
                        copy_protection['domain_editors'] = protected_range['editors']['domain_users_can_edit']

                if 'start_row_index' in protected_range['range']:
                    copy_protection['start_row'] = protected_range['range']['start_row_index']
                if 'start_column_index' in protected_range['range']:
                    copy_protection['start_column'] = protected_range['range']['start_column_index']
                if 'end_row_index' in protected_range['range']:
                    copy_protection['last_row'] = protected_range['range']['end_row_index']-1
                if 'end_column_index' in protected_range['range']:
                    copy_protection['last_column'] = protected_range['range']['end_column_index']-1

                if 'unprotected_ranges' in protected_range:
                    copy_protection['unprotected_ranges'] = [
                        {
                            'row':  upr['start_row_index'],
                            'column':  upr['start_column_index'],
                            'last_row':  upr['end_row_index']-1,
                            'last_column':  upr['end_column_index']-1,
                        } for upr in protected_range['unprotected_ranges']
                    ]
                new_ws.protect(**copy_protection)
            new_ws.commit()

        if title:
            new_ws.title = title
            new_ws.commit()
        elif keep_title:
            new_ws.title = self.title
            new_ws.commit()
            #spreadsheet.commit()
        return(new_ws)

    def record_value_change(self, property, cleared=False):
        if cleared:
            del self.changes[property]
        elif property == 'tab_color':
            raise Exception("Not handled yet")
        elif property in ('frozen_row_count', 'frozen_column_count', 'hide_gridlines'):
            camelprop = googleapps.toCamel(property)
        elif property == 'rows_available':
            camelprop = 'row_count'
        elif property == 'columns_available':
            camelprop = 'column_count'

            self.batch_update(
                'updateSheetProperties',
                properties={'sheetId': lambda s: s.id, 'gridProperties': {camelprop, getattr(self, property)} },
                fields='gridProperties.' + camelprop,
            )
            raise Exception("Not handled yet")
        else:
            camelprop = googleapps.toCamel(property)
            self.batch_update(
                'updateSheetProperties',
                properties={'sheetId': lambda s: s.id, camelprop: getattr(self, property)},
                fields=camelprop,
            )

    def __init__(self, spreadsheet, json=None, **default_params):
        super().__init__()
        self.spreadsheet = spreadsheet

        for property, val in default_params.items():
            if property not in self.api_properties + self.api_ro_properties:
                raise Exception("Unknown parameter " + property)
            setattr(self, '_' + property, val)

        if json:
            self.process_json(json)

        self.protected_requests = []
        self._celldata = None
        self._celldata_type = None
        self._ncolumns = None
    #def process_spreadsheet_json_response(self, json):
        # this function is primarily used to extract the sheet-specific json from a full response from a spreadsheet get/create call
        #

    def process_new_worksheet_json(self, json):
        # this is broken as it doesn't allow for multiple batches
        return(self.process_json(json['replies'][0]['addSheet']))

    def process_json(self, json):

        # Note/TODO: This doesn't provide support to clear a field by setting to None.  Is that an issue?
        # -- first handle the properties
        json = googleapps.unCamel(json)
        self._original_json = json

        #-- first do properties
        params = googleapps.unCamel(json['properties'], flatten=True)

        self._id = params['sheet_id']
        self._columns_available = params['column_count']
        self._rows_available = params['row_count']
        for field in ('frozen_row_count', 'frozen_column_count', 'title', 'index'):
            if field in params and params[field] is not None:
                setattr(self, '_' + field, params[field])
            #TODO: unknown parameters are not being caught and killed
            # therea re at least red,green,blue that need to be accounted for
            # herd
            # if property not in self.api_properties:
            #    print("TODO: Ignoring unknown parameter " + param)
            #if params[param] is not None:
            #    setattr(self, '_' + property, params[param])

        # -- other core fields
        if 'merges' in json:
            self._merges = json['merges']
        if 'protected_ranges' in json:
            self._protected_ranges = json['protected_ranges']
        else:
            self._protected_ranges = None
        # -- then the data
        if 'data' in json:
            self._celldata = json['data'][0]
            self._celldata_type = 'grid'

    def process_spreadsheet_values_json(self, json):
        json = googleapps.unCamel(json)
        if json['major_dimension'] != 'ROWS':
            raise Exception("no logic to process columns yet")
        if 'values' in json:
            self._celldata = json['values']
        else:
            self._celldata = []
        self._celldata_type = 'values'

    @property
    def celldata_type(self):
        if not self._celldata_type:
            self.celldata
        return(self._celldata_type)

    @property
    def celldata(self):
        if self._celldata:
            return(self._celldata)

        # make the call
        # this can be a problem if the rows_available or columns_available is out of sync with the current sheet

        self.commit() #-- this is necessary in case we are refreshing.
        self.make_request(['spreadsheets', 'values'], 'get',
                          spreadsheetId=self.spreadsheet.id,
                          range=googleapps.sheets.range_to_a1(0,0,self.rows_available-1, self.columns_available-1, worksheet=self),
                          callback=self.process_spreadsheet_values_json
        )
        return(self._celldata)

    def refresh(self):
        self._celldata = None
        return(self.celldata)

    @property
    def protected_ranges(self):
        return(self._protected_ranges)


    def get_values(self):
        if self.celldata_type == 'values':
            # should this be a copy?
            return(copy.deepcopy(self.celldata))

        raise Exception("TODO: get_values from grid data")

    @sheetrange_method
    def get_range(self, start_row, start_col, end_row=None, end_col=None):

        if end_row is None:
            end_row = start_row
        if end_col is None:
            end_col = start_col

        result_data = []

        if self.celldata_type == 'values':

            rowdata = []
            for irow in range(start_row, end_row+1):
                if irow >= len(self.celldata):
                    break
                row = []
                for icol in range(start_col, end_col+1):
                    if icol >= len(self.celldata[irow]):
                        break
                    row.append(self.celldata[irow][icol])
                rowdata.append(row)
            return(rowdata)


        raise Exception("NEed to re-evaluate how ranges are done for grid data")

        for irow in range(start_row, end_row+1):

            if irow >= len(self.celldata['row_data']):
                # no more data
                # TODO: Verify this is the correct assumption
                # if it's the end of the sheet, we don't append blank rows
                # we could append(None) to have a matrix exactly as long as requested
                break
            for icol in range(start_col, end_col+1):
                if 'values' not in self._celldata['row_data'][irow]:
                    # this is a blank row.
                    rowdata.append(None)
                elif icol >= len(self._celldata['row_data'][irow]['values']):
                    # TODO: Verify this is the correct assumption
                    # same as the prior: if this is a column request past the point with information, we just break out
                    # and don't pad the array. This can lead to rows that have different lengths
                    # Alternatively, we could append(None) to have a matrix exactly as long as requested
                    break
                elif not 'effective_value' in self._celldata['row_data'][irow]['values'][icol]:
                    rowdata.append(None)
                else:
                    celldata = self._celldata['row_data'][irow]['values'][icol]['effective_value']

                    if 'string_value' in celldata:
                        rowdata.append(celldata['string_value'])
                    elif 'number_value' in celldata:
                        rowdata.append(celldata['number_value'])
                    elif 'boolean_value' in celldata:
                        rowdata.append(celldata['boolean_value'])
                    else:
                        print("Failed to append (" + str(irow) + "," + str(icol) + ")")
                        print('Cell data is :', celldata)
                        raise Exception("don't know the data in this cell!")

            result_data.append(rowdata)
        return(result_data)

    @sheetrange_method
    def get_row(self, *args):
        data = self.get_range(*args)
        if len(data) > 1:
            raise Exception("get_rows should be provided with a range that is just one row")
        return(data[0])

    @rowcol_method
    def get_cell(self, row, col):
        data = self.get_range(row,col,row,col)
        if len(data) > 1 or len(data[0]) > 1:
            raise Exception("get_cell should be provided with a range that is just one cell")
        if not data or not data[0]:
            return(None)
        return(data[0][0])





    def batch_update(self, request_type, **parameters):
        # this is an encapsulation failure
        # TODO: let the singleton service queue up the requests in order

        if self.last_request and self.last_request.function == 'batchUpdate' and not self.debug_commit:
            self.last_request.append('body', 'requests', {request_type: parameters})
        else:
            self.add_request('spreadsheets', 'batchUpdate',
                spreadsheetId=lambda s: s.spreadsheet.id,
                body={
                    'requests': [ {request_type: parameters} ]
                })

    def update_row(self, row, *args, **kwargs):
        self.update_dimension_properties('rows', row, row, *args, **kwargs)

    def update_rows(self, start_row, end_row, *args, **kwargs):
        self.update_dimension_properties('rows', start_row, end_row, *args, **kwargs)

    def update_column(self, column, *args, **kwargs):
        self.update_dimension_properties('columns', column, column, *args, **kwargs)

    def update_columns(self, start_column, end_column, *args, **kwargs):
        self.update_dimension_properties('columns', start_column, end_column, *args, **kwargs)

    def update_dimension_properties(self, dimension_type, start, end, size=None, hidden=None):
        properties = {}
        if size:
            properties['pixelSize'] = size
        if hidden is not None:
            properties['hiddenByUser'] = hidden

        if not properties:
            raise Exception("update dimension called for " + dimension_type + ", but no properties set/changed")

        self.batch_update(
            'updateDimensionProperties',
            range=self.dimension_range(dimension_type, start, end),
            properties=properties,
            fields='*',
        )


    def append_rows(self, count):
        return(self.append_dimension('rows', count))

    def append_columns(self, count):
        return(self.append_dimension('columns', count))

    def append_dimension(self, dimension_type, count):
        self.batch_update(
            'appendDimension',
            sheetId=lambda s: s.id,
            dimension=dimension_type.upper(),
            length=count
        )
        if dimension_type.lower() == 'rows':
            self._rows_available += count
        elif dimension_type.lower() == 'columns':
            self._columns_available += count
        else:
            raise Exception("Unknown dimension " + str(dimension_type))

    def delete_row(self, row): return(self.delete_dimension('ROWS', row))
    def delete_rows(self, start, end=None): return(self.delete_dimension('ROWS', start, end))
    def delete_column(self, column): return(self.delete_dimension('COLUMNS', column))
    def delete_columns(self, start, end=None): return(self.delete_dimension('COLUMNS', start, end))

    def delete_dimension(self, dimension_type, start, end=None):
        self.batch_update(
            'deleteDimension',
            range=self.dimension_range(dimension_type, start, end)
        )
        if end:
            count = end - start
        else:
            count = 1

        if dimension_type.lower() == 'rows':
            self._rows_available -= count
            # if this overlaps with rows with data, we need to update that too
            if start < self.nrows:
                # we need to remove these from the celldata array
                if start + count < self.nrows:
                    del self.celldata[start:(start+count)]
                else:
                    del self.celldata[start:self.nrows]
        elif dimension_type.lower() == 'columns':
            self._columns_available -= count

            # if ncolumns has been calculated, adjust it too
            if start < self.ncolumns:
                raise Exception('TODO: handle column deletes when data is present')
        else:
            raise Exception("Unknown dimension " + str(dimension_type))

    def _addif(self, ref, param, data):
        if param in data and data[param] is not None:
            ref[googleapps.toCamel(param)] = data[param]
            del data[param]



    def boolean_condition(self, condition_type, values=None):
        translate_condition = {
            '>': 'NUMBER_GREATER',
            '>=': 'NUMBER_GREATER_THAN_EQ',
            '<': 'NUMBER_LESS',
            '<=': 'NUMBER_LESS_THAN_EQ',
            '==': 'NUMBER_EQ',
            #HACK
            #            '=': 'NUMBER_EQ',
            '=': 'TEXT_EQ',
            '!=':'NUMBER_NOT_EQ',
            '<>':'NUMBER_NOT_EQ',
            'BETWEEN': 'NUMBER_BETWEEN',
            'NOT_BETWEEN': 'NUMBER_NOT_BETWEEN',
            'CONTAINS': 'TEXT_CONTAINS',
            'NOT_CONTAINS': 'TEXT_NOT_CONTAINS',
            'STARTS_WITH': 'TEXT_STARTS_WITH',
            'ENDS_WITH': 'TEXT_ENDS_WITH',
            'EQ': 'TEXT_EQ',
            'EMAIL': 'TEXT_IS_EMAIL',
            'URL': 'TEXT_IS_URL',
            'DATE_EQ': 'DATE_EQ',
            'DATE_BEFORE': 'DATE_BEFORE',
            'DATE_AFTER': 'DATE_AFTER',
            'DATE_ON_OR_BEFORE': 'DATE_ON_OR_BEFORE',
            'DATE_ON_OR_AFTER': 'DATE_ON_OR_AFTER',
            'DATE_BETWEEN': 'DATE_BETWEEN',
            'DATE_NOT_BETWEEN': 'DATE_NOT_BETWEEN',
            'DATE_IS_VALID': 'DATE_IS_VALID',
            'ONE_OF_RANGE': 'ONE_OF_RANGE',
            'ONE_OF_LIST': 'ONE_OF_LIST',
            'BLANK': 'BLANK',
            'NOT_BLANK': 'NOT_BLANK',
            'CUSTOM_FORMULA': 'CUSTOM_FORMULA',
        }

        condition = {}
        if values is not None:
            condition['values'] = list(map(lambda v:{'userEnteredValue':v}, values))
        # now for conditions
        condiition_type = condition_type.upper()
        if condition_type in translate_condition:
            condition['type'] = translate_condition[condition_type]
        else:
            raise Exception("Unknown condition type '" + condition_type + "'")
        return(condition)

    @sheetrange_method
    def add_conditional_formatting(self, start_row, start_column, last_row, last_column, rule_type, format, values=None):

        #TODO: GradientRule
        self.batch_update(
            'addConditionalFormatRule',
            rule={
                'ranges': [self.grid_range(start_row, start_column, last_row, last_column)],
                'booleanRule': {
                    'condition': self.boolean_condition(rule_type, values),
                    'format': format.cellformat_json(),
                },
            },
        )

    @sheetrange_method
    def add_data_validation(self, start_row, start_column, last_row, last_column, validation_type, values=None, required=True, input_message=None, show_dropdown=False):
        request = {
            'strict': required,
            'showCustomUi': show_dropdown,
            'condition': self.boolean_condition(validation_type, values),
        }

        if input_message is not None:
            request['inputMessage'] = input_message

        self.batch_update(
            'setDataValidation',
            range=self.grid_range(start_row, start_column, last_row, last_column),
            rule=request,
        )

    @sheetrange_method
    def unmerge_cells(self, start_row, start_column, last_row, last_column):
        self.batch_update(
            'unmergeCells',
            range=self.grid_range(start_row, start_column, last_row, last_column)
        )

    @sheetrange_method
    def merge_cells(self, start_row, start_column, last_row, last_column, merge_type='merge_all'):
        merge_type = merge_type.upper()
        if merge_type and merge_type not in ('MERGE_ALL', 'MERGE_ROWS', 'MERGE_COLUMNS'):
            raise Exception("Unknown merge_type '"+merge_type+"'")

        self.batch_update(
            'mergeCells',
            mergeType=merge_type,
            range=self.grid_range(start_row, start_column, last_row, last_column)
        )

    def freeze_rows(self, number_frozen):
        self.batch_update(
            'updateSheetProperties',
            properties={
                'sheetId': lambda s: s.id,
                'gridProperties': {'frozenRowCount': number_frozen},
                },
            fields='gridProperties.frozenRowCount'
        )

    def freeze_columns(self, number_frozen):
        self.batch_update(
            'updateSheetProperties',
            properties={
                'sheetId': lambda s: s.id,
                'gridProperties': {'frozenColumnCount': number_frozen},
            },
            fields='gridProperties.frozenColumnCount'
        )

    def freeze_panes(self, rows_frozen, columns_frozen):
        self.batch_update(
            'updateSheetProperties',
            properties={
                'sheetId': lambda s: s.id,
                'gridProperties': {
                    'frozenColumnCount': columns_frozen,
                    'frozenRowCount': rows_frozen},
            },
            fields='gridProperties.frozenColumnCount,gridProperties.frozenRowCount'
        )

    @sheetrange_method
    def autofilter(self, row=0, column=0, last_row=None, last_column=None, sort_specification=None, criteria=None):
        """ Apply the 'Basic Filter' to the sheet with the specified range.

        Args:
            range {GridRange}: the area to convert into a filter. By default, A1 to the last row and column. User can specify any set of parameters and the default will fill in the rest.
            sort_specification {dict} [optional]: a SortSpec json object. See https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#BasicFilter
            criteria {dict} [optional]: a FilterCriteria json object. See https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#BasicFilter
         """
        if last_row is None:
            last_row = self.nrows - 1
        if last_column is None:
            last_column = self.ncolumns - 1

        self.batch_update(
            'setBasicFilter',
            filter={
                'range':self.grid_range(row, column, last_row, last_column),
                'sortSpecs': sort_specification,
                'criteria': criteria,
            }
        )

    def clear_autofilter(self):
        """ Clears the existing autofilter. A sheet can only have on 'Basic Filter', so no parameters are needed. """

        self.batch_update(
            'clearBasicFilter',
            sheetId=lambda s: s.id,
        )


    @sheetrange_method
    def sort_range(self, row, column, last_row, last_column, *sort_specifications):
        """ sort a range in the spreadsheet.  sort specifications should be an array as follows:

            Empty array: sort the first column in the range, ascending
            {int}: A simple specification to indicate the column_offset to sort in ascending order
            {object of {'coloffset': int, 'reverse': bool}} reverse  is technically optional"""


        if not sort_specifications:
            specification = [{'dimensionIndex': 0}]
        else:
            specification = []
            for coldef in sort_specifications:
                if type(coldef) is int:
                    specification.append({'dimensionIndex': coldef, 'sortOrder': 'ASCENDING'})
                else:
                    sortdef = {'dimensionIndex': coldef['coloffset'], 'sortOrder': 'ASCENDING'}
                    if 'reverse' in coldef and coldef['reverse']:
                        sortdef['sortOrder'] = 'DESCENDING'
                    specification.append(sortdef)


        self.batch_update(
            'sortRange',
            range=self.grid_range(row, column, last_row, last_column),
            sortSpecs=specification
        )
        #
        # TODO: the result value returns the reordered data, so we should incorporate that more elegantly
        #
        self._celldata = None

    def protect(self, start_row=None, start_column=None, last_row=None, last_column=None, user_editors=None, group_editors=None, description=None, warning_only=False, unprotected_ranges=None, domain_editors=None):
        request = {'warningOnly':warning_only}
        if description:
            request['description'] = description

        if user_editors or group_editors:
            request['editors'] = {}
            if user_editors: request['editors']['users'] = user_editors
            if group_editors: request['editors']['groups'] = group_editors
        else:
            # TODO: the editors field needs to be restricted somehow, otherwise anyone with whom it is shared
            # can edit... Which is not what most people expect.
            # So a more elegant solution needs to be foound. Maybe adding self explicilty as the only editor?
            request['editors'] = {'domainUsersCanEdit':False}

        if domain_editors:
            request['editors']['domainUsersCanEdit'] = True

        if start_row is None and start_column is None and last_row  is None and last_column is None:
            #- sheet level protections
            request['range'] = {'sheetId': lambda s: s.id}
        elif start_row is not None and last_row:
            if start_column is not None and last_column:
                # full range
                request['range'] = self.grid_range(start_row, start_column, last_row, last_column)
            else:
                # just a set of rows
                request['range'] = self.grid_range(start_row, start_column, last_row, last_column)
        elif start_column is not None and last_column:
            # just a set of columns
            request['range'] = self.grid_range(start_row, start_column, last_row, last_column)

        else:
            raise Exception("Grid range not complete!")

        if unprotected_ranges:
            request['unprotectedRanges'] = list(map(lambda up: self.grid_range(**up), unprotected_ranges ))

        # TODO: what about domainUsersCanEdit
        # we can have multiple protected ranges, but only one sheet-level protection.
        # So die if that happens

        if 'startRowIndex' not in request['range'] and self.protected_requests:
            for prior in self.protected_requests:
                if 'startRowIndex' not in prior['request']['range']:
                    raise Exception("Can only protect a sheet one time")

        self.protected_requests.append({'request': request, 'type': 'new', 'unprotect': []})

        #if self.last_request and self.last_request.function == 'batchUpdate' and not self.debug_commit:
        #    self.last_request.append('body', 'requests', {request_type: parameters})
        #else:
        #self.batch_update('addProtectedRange',protectedRange=request)

    @sheetrange_method
    def unprotect(self, start_row, start_column, last_row, last_column, user_editors=None, group_editors=None, description=None, warning_only=False):

        request = {}
        # TODO: This is currently limited to sheets that have already been protected.
        # Also need to verify whether this is additive
        for protected_request in self.protected_requests:
            if 'start_row_index' not in protected_request['request']['range'] or (
                    # see if the unprotected request falls within here
                    start_row >= protected_request['request']['range']['start_row_index']
                    and last_row <= protected_request['request']['range']['end_row_index']
                    and start_column >= protected_request['request']['range']['end_column_index']
                    and last_column <= protected_request['request']['range']['end_column_index']
                    ):

                # TODO: this doesn't handle the istuation where we add a new protected range
                # and then try to unprotect cells in a different range
                protected_request['unprotect'].append(self.grid_range(start_row, start_column, last_row, last_column))
                return(True)
        else:
            # check to see if adding editors to an an existing protected range
            # if it is, we'll convert it to a request
            for protected in self.protected_ranges:
                if 'start_row_index' not in protected['range'] or (
                    # see if the unprotected request falls within here
                    start_row >= protected['range']['start_row_index']
                    and last_row <= protected['range']['end_row_index']
                    and start_column >= protected['range']['end_column_index']
                    and last_column <= protected['range']['end_column_index']
                    ):
                    #--- it would appear that this unprotected request
                    # falls within this range
                    # so we'll add the update request
                    # and return
                    if 'unprotected_ranges' in protected:
                        unprotected_ranges = protected['unprotected_ranges']
                    else:
                        unprotected_ranges = []

                    # add it here
                    unprotected_ranges.append(self.grid_range(start_row, start_column, last_row, last_column))

                    self.protected_requests.append({
                        'type': 'update',
                        'request': protected,
                        'unprotect': unprotected_ranges,
                    })
                    return(True)

        # if we got here, we couldn't find a protected range
        # that matches here
        raise Exception("Sheet not protected. Cannot add uprotected section.")

    #    if unprotected_ranges:
    #        request['unprotectedRanges'] = list(map(lambda up: self.grid_range(**up), unprotected_ranges ))

        # TODO: what about domainUsersCanEdit
#        self.batch_update('addProtectedRange', protectedRange=request, fields='unprotectedRanges')


    def commit(self):
        # if there is a protection request, add it
        if self.protected_requests:
            for protected_request in self.protected_requests:
                request = protected_request['request']
                if len(protected_request['unprotect']):
                    request['unprotectedRanges'] = protected_request['unprotect']

                if protected_request['type'] == 'new':
                    self.batch_update('addProtectedRange', protectedRange=request)

                elif protected_request['type'] == 'update':
                    self.batch_update(
                        'updateProtectedRange',
                        protectedRange={
                            'protectedRangeId': protected_request['request']['protected_range_id'],
                            'unprotectedRanges': protected_request['unprotect'],
                        },
                        fields='unprotectedRanges',
                    )

            # clear this out so it doesn't get processed a second time on a second commit
            self.protected_requests = []
        try:
            result = super().commit()
        except googleapiclient.errors.HttpError as he:
            # catch and rethrow certain errors
            import sys
            exc_class, exc, tb = sys.exc_info()
            reason = he._get_reason()

            #
            # Recode errors that exceed the width or height of the spreadsheet to IndexError
            #
            update_out_of_bounds = re.match('Invalid requests\[\d+\].updateCells: GridCoordinate.(\w\w\w)\w*\[(\d+)\] is after last (row|col) in grid\[(\d+)\]', reason)
            if update_out_of_bounds:
                raise IndexError("Attempt to commit updates to Google Spreadsheet '"+self.spreadsheet.title+"', Worksheet '"+self.title+"': Cell specified to update in " + update_out_of_bounds.group(1) + " " + str(update_out_of_bounds.group(2)) + ", which is after the last "+update_out_of_bounds.group(3)+" ("+update_out_of_bounds.group(4)+")") from he

            #
            # default is to also modify the error message to add in information about the spreadsheet and worksheet
            #
            # This may be misguided, but I find the default HttpError messages to not be useful for large automations,
            # So I'm adding details here to at least reference the worksheet by name.
            #
            import json
            jsoncontent = json.loads(he.content)
            jsoncontent['error']['message'] += " (error occurred when commiting changes to Spreadsheet '"+self.spreadsheet.title+"', Worksheet '"+self.title+"')"
            he.content = json.dumps(jsoncontent).encode('utf-8')
            raise

        return(result)


    @property
    def nrows(self):
        if self.celldata is None:
            return(0)
        return(len(self.celldata))

    @property
    def ncolumns(self):
        # check if ncolumns is cached OR if there are no rows
        if self._ncolumns is not None or not self.nrows:
            return(self._ncolumns)

        self._ncolumns = 0
        for i in range(len(self.celldata[0])):
            if len(self._celldata[i]) > self._ncolumns:
                self._ncolumns = len(self._celldata[i])
        return(self._ncolumns)

    def insert_rows(self, index, count, data_matrix=None, inherit_prior_style=True, format=None):
        self.insert_dimensions('ROWS', index, count, data_matrix, inherit_prior_style, format)

    def insert_row(self, index, values=None, inherit_prior_style=True, format=None):
        self.insert_dimensions('ROWS', index, 1, values, inherit_prior_style, format)

    def insert_column(self, index, values=None, inherit_prior_style=True, format=None):
        self.insert_dimensions('COLUMNS', index, 1, values, inherit_prior_style, format)

    def insert_columns(self, index, count, data_matrix=None, inherit_prior_style=True, format=None):
        self.insert_dimensions('COLUMNS', index, count, data_matrix, inherit_prior_style, format)

    def insert_dimensions(self, dimension_type, index, count, value_matrix=None, inherit_prior_style=True, format=None):

        dimension_type = dimension_type.upper()

        if dimension_type == 'ROWS':
            if index > self.nrows:
                raise Exception("Attempt to insert new rows at index " + str(index) + " cannot be done because the worksheet does not have " + str(index) + " rows to begin with!")
        elif dimension_type == 'COLUMNS':
            if index > self.ncolumns:
                raise Exception("Attempt to insert new columns at index " + str(index) + " cannot be done because the worksheet does not have " + str(index) + " columns to begin with!")
        else:
            raise Exception("Unknown dimension " + str(dimension_type))

        self.batch_update(
            'insertDimension',
            range={
                'sheetId':lambda s: s.id,
                'dimension':dimension_type,
                'startIndex': index,
                'endIndex': index + count,
            },
            inherit_from_before=inherit_prior_style,
        )

        if dimension_type == 'ROWS':
            self._rows_available += count
        elif dimension_type == 'COLUMNS':
            self._columns_available += count
        else:
            raise Exception("Unknown dimension " + str(dimension_type))


        if value_matrix or format:
            self.update_cells(index, 0, value_matrix, format)


    def append(self, row, format=None):
        """ adds a single row to the end of the worksheet."""
        # last_row is not a variable, so we just need to add it to celldata

        if type(row[0]) in (list, tuple):
            raise Exception("Should not append a matrix - append is for a single row at a time. Use Extend")

        # add more rows if we have exceeded the sheet
        if self.nrows + 1 >= self.rows_available:
            self.append_rows(250)

        result = self.update_cells(self.nrows, 0, row, format=format)
        self._celldata.append(row)
        return(result)


    def extend(self, rows, format=None):
        """ adds a set of rows to the end of the worksheet."""
        # last_row is not a variable, so we just need to add it to celldata

        if type(rows[0]) not in (list, tuple):
            raise Exception("extend should be a matrix - e.g. a list of lists.")

        # add more rows if we have exceeded the sheet
        if self.nrows + len(rows) >= self.rows_available:
            nexthundred = (int((self.nrows + len(rows) - self.rows_available) / 100) + 2) * 100
            self.append_rows(nexthundred)

        result = self.update_cells(self.nrows, 0, rows, format=format)
        self._celldata.extend(rows)
        return(result)


    @sheetrange_method
    def update_format(self, row, column, last_row, last_column, cellformat):
        #if 'right_border' in cellformat:
        #    import pdb
        #    pdb.set_trace()

        if type(cellformat) is dict:
            cellformat = googleapps.sheets.Format(**cellformat)

        # normal formatting

        format_json = cellformat.cellformat_json()

        prepared_rows = []
        for valuerow in range(row,last_row+1):
            prepared_rows.append({'values': [ {'userEnteredFormat': format_json} for j in range(column, last_column+1) ]})

        fields = cellformat.fieldmask_fields()
        self.batch_update('updateCells',
                          rows=prepared_rows,
                          start=self.grid_coordinate(row, column),
                          #range=self.grid_range(row, column, last_row, last_column),
                          fields=','.join(['userEnteredFormat.' + f for f in fields]),
                          )

    @rowcol_method
    def update_cells(self, row, column, values=None, format=None):

        request = {}
        fields_active = ['userEnteredValue']
        #-----------------------------------
        # Process Formatting

        if format:
            if type(format) is dict:
                format = googleapps.sheets.Format(**format)
            # normal formatting
            format_json = format.cellformat_json()
            if format_json:
                fields_active.extend(googleapps.json_fieldmask({'userEnteredFormat':format_json}))
        else: format_json = None

        #-----------------------------------
        # TODO: Data Validation!
        # We can add it here!
        #

        #-----------------------------------
        # Actual Data! get everything into matrix format!
        if type(values) not in (list,tuple):
            values = [ values ]
        if len(values) and type(values[0]) not in (list,tuple):
            values = [ values ]

        prepared_rows = []
        for valuerow in values:
            prepared_row = []
            for cell in valuerow:
                # single cell value
                if type(cell) is dict:
                    prepared_row.append(cell)
                else:
                    celljson = {'userEnteredValue': self.extended_value(cell) }
                    if format_json:
                        celljson['userEnteredFormat'] = format_json
                    prepared_row.append(celljson)
            prepared_rows.append({'values': prepared_row})


        self.batch_update('updateCells',
                          start=self.grid_coordinate(row, column),
                          rows=prepared_rows,
                          fields=','.join(fields_active) # TODO < this isn't robust
        )

        '''
        don't do this now
        # now process the range/cell protection
        # it may eventually be more efficient to have this done after the fact, at the end of batch processing
        # but for now, we do it real time
        if format and format.locked is not None:
            meat = {}
            if format.locked:
                meat = {'protectedRange':self.grid_range(row, col, row+len(rows), col+len(rows[0]))}
            else:
                meat = {'unprotectedRanges':[self.grid_range(row, col, row+len(rows), col+len(rows[0]))]}

            self.services.add_request('spreadsheets', 'batchUpdate', {
            'spreadsheetId': self.workbook.id,
            'body': {
                'addProtectedRange': meat,
                }
            })
        '''

    @classmethod
    def extended_value(self, cell_value):

        if cell_value is dict or cell_value is None:
            # we assume they already processed this
            # None is the same as blank, won't update anything
            return(cell_value)

        # okay, it's something
        t = type(cell_value)
        if t is str:
            # is it a formula?
            if len(cell_value) and cell_value[0] == '=':
                return({'formulaValue': cell_value})
            else:
                return({'stringValue': cell_value})

        elif t is int or t is float:
            return({'numberValue': cell_value})

        elif t is bool:
            return({'boolValue': cell_value})

        # todo: should this be rendered as a number in SERIAL__NUMBER format
        # from the docs: The whole number portion of the value (left of the decimal) counts the days since December 30th 1899. The fractional portion (right of the decimal) counts the time as a fraction of the day. For example, January 1st 1900 at noon would be 2.5
        elif t is datetime.date:
            diff = cell_value - serial_number_base.date()
            return({'numberValue': diff.days})
        elif t is datetime.datetime:
            diff = cell_value - serial_number_base
            return({'numberValue': diff.total_seconds() / 60 / 60 / 24})

        raise TypeError("Unknown type '"+str(t)+"' cannot be processed by update cells for this value: " + str(cell_value))


    @sheetrange_method
    def grid_range(self, row, column, last_row, last_column):
        return ({
                'sheetId': lambda s: s.id,
                "startRowIndex": row,
                "endRowIndex": last_row+1 if last_row is not None else None,
                "startColumnIndex": column,
                "endColumnIndex": last_column+1 if last_column is not None else None,
        })

    @rowcol_method
    def grid_coordinate(self, row,column):
        return ({
                'sheetId': lambda s: s.id,
                'rowIndex': row,
                'columnIndex': column,
        })

    def dimension_range(self, type, start, end=None):
        type = type.upper()
        if type not in ('ROWS', 'COLUMNS'):
            raise Exception("Dimension type '" + type + "' is unknown")

        if not end:
            end=start
        return ({
                'sheetId': lambda s: s.id,
                'dimension': type,
                'startIndex': start,
                'endIndex': end+1,
        })



# Copyright (c) 2019-2021 Kevin Crouse
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @license: http://www.apache.org/licenses/LICENSE-2.0
# @author: Kevin Crouse (krcrouse@gmail.com)

import re
import warnings
from pprint import pprint
import googleapps
from googleapps.serviceclient import CommitType

@googleapps.serviceclient.googleappclient
class Spreadsheet(googleapps.ServiceClient):
    default_api = 'sheets'
    default_api_version = 'v4'
    api_ro_properties = ('id',)
    api_properties = ('title', 'locale', 'auto_recalc', 'time_zone', 'default_format')
    commit_policy=CommitType.defer_commit


    def __init__(self):
        super().__init__()
        self._worksheets = []


    def commit(self):
        # TODO : let the singleton service handle commit.
        # This has problems with sync issues - what if spreadsheet things change
        # that are partly dependent on the underlying worksheettchanges?
        if self.debug:
            print("---------< COMMIT COMMIT COMMIT ")
        super().commit()
        if self.worksheets:
            for worksheet in self.worksheets:
                if self.debug:
                    print(" --------< Recursive Commit for Sheet " + worksheet.title)
                worksheet.commit()

    @property
    def active(self):
        return(self._worksheets[0])

    def add_worksheet(self, title=None, **kwargs):
        #we do the title like this so we can pass it in as an array  obj.add_worksheet("label")
        kwargs['title'] = title
        '''        json_definition = ('title', 'index', 'hidden', 'tab_color', 'right_to_left', 'sheet_id', 'sheet_type',)
        grid_properties = ('row_count', 'column_count', 'frozen_row_count', 'frozen_column_count','hide_gridlines',)
        request = self.process_api_definition('add_worksheet',
                                              kwargs,
                                              json_definition,
                                              {'grid_properties': grid_properties})
        '''
        new_worksheet = self._add_uncommited_worksheet(**kwargs)
        #TODO: This interface is not complete
        self.add_request('spreadsheets', 'batchUpdate',
                         spreadsheetId=lambda s: s.id,
                         body={
                             'requests': [{'addSheet': {'properties': {'title': new_worksheet.title }}}]
                         },
                         callback=new_worksheet.process_new_worksheet_json
                         )
        return(new_worksheet)

    def _add_uncommited_worksheet(self, **kwargs):
        new_worksheet = googleapps.sheets.Worksheet(self, rows_available=1000, columns_available=26, **kwargs)
        self._worksheets.append(new_worksheet)
        return(new_worksheet)

    def get_worksheet(self, title):
        for worksheet in self._worksheets:
            if worksheet.title == title:
                return(worksheet)
        return

    @property
    def url(self):
        return(self._url)

    @property
    def worksheets(self):
        return(self._worksheets)

    @property
    def worksheet_titles(self):
        return([ws.title for ws in self.worksheets])

    def batch_update(self, request_type, **parameters):
        if self.last_request and self.last_request.function == 'batchUpdate':
            self.last_request.append('body', 'requests', {request_type: parameters})
        else:
            self.add_request('spreadsheets', 'batchUpdate',
                spreadsheetId=lambda s: s.id,
                body={
                    'requests': [ {request_type: parameters} ]
                })


    def process_json(self, spreadsheet_json):
        self._id  = spreadsheet_json['spreadsheetId']
        self._url  = spreadsheet_json['spreadsheetUrl']
        for param in self.api_properties:
            setattr(self, '_' + param, spreadsheet_json['properties'][googleapps.toCamel(param)])

        # Process sheets
        for worksheetjson in spreadsheet_json['sheets']:
            # see if the worksheet already exists
            worksheet = self.get_worksheet_by_id(worksheetjson['properties']['sheetId'])
            if worksheet:
                # update anything that needs it
                worksheet.process_json(worksheetjson)
            else:
                # change name? Actually commited!
                worksheet = self._add_uncommited_worksheet(json=worksheetjson)
        # spreadsheet_json['sheets']
        return

    def get_worksheet_by_id(self, id):
        # encapsulated in case we ever want to move to a hashmap store
        for worksheet in self.worksheets:
            if worksheet.id == id:
                return(worksheet)
        return
